import warnings
import numpy as np
import networkx as nw
import unittest
from unittest.mock import Mock
import warnings
import os

import numpy as np
from scipy.optimize._minimize import MINIMIZE_METHODS

from openqaoa.qaoa_components import create_qaoa_variational_params, QAOADescriptor, PauliOp, Hamiltonian
from openqaoa.utilities import X_mixer_hamiltonian
from openqaoa.backends.qaoa_backend import get_qaoa_backend
from openqaoa.backends import create_device
from openqaoa.optimizers import get_optimizer
from openqaoa.algorithms.qaoa.qaoa_result import QAOAResult
from openqaoa.derivatives.derivative_functions import derivative
from openqaoa.optimizers.logger_vqa import Logger
from openqaoa.derivatives.qfim import qfim
from openqaoa.problems import MinimumVertexCover
"""
Unittest based testing of custom optimizers.
"""

cost_hamiltonian_1 = Hamiltonian([PauliOp('ZZ', (0, 1)), PauliOp('ZZ', (1, 2)), PauliOp('ZZ', (0, 3)), PauliOp('Z', (2,)), PauliOp('Z', (1,))], [1, 1.1, 1.5, 2, -0.8], 0.8)

class TestQAOACostBaseClass(unittest.TestCase):
    
    def setUp(self):
        
        self.log = Logger({'func_evals': 
                           {
                               'history_update_bool': False, 
                               'best_update_string': 'HighestOnly'
                           },
                           'jac_func_evals':
                           {
                               'history_update_bool': False, 
                               'best_update_string': 'HighestOnly'
                           },
                           'qfim_func_evals': 
                           {
                               'history_update_bool': False, 
                               'best_update_string': 'HighestOnly'
                           }
                          }, 
                          {
                              'root_nodes': ['func_evals', 'jac_func_evals', 
                                             'qfim_func_evals'], 
                              'best_update_structure': []
                          })
        
        self.log.log_variables({'func_evals': 0, 'jac_func_evals': 0, 'qfim_func_evals': 0})

    def __backend_params(self, cost_hamil, n_qubits):
        """ Helper function to create a backend and a parameters onjects for testing. """
        mixer_hamil = X_mixer_hamiltonian(n_qubits=n_qubits)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=1)
        device = create_device('local','vectorized')
        backend_obj_vectorized = get_qaoa_backend(qaoa_descriptor,device)
        variate_params = create_qaoa_variational_params(qaoa_descriptor, 'standard', 'ramp')
        return backend_obj_vectorized, variate_params
        
    def test_saving_feature(self):
        """ Test save_intermediate in OptimizeVQA """
        backend_obj_vectorized, variate_params = self.__backend_params(cost_hamiltonian_1, 4)        
        niter = 5
        grad_stepsize = 0.0001
        stepsize = 0.001

        params_array = variate_params.raw().copy()
        jac = derivative(backend_obj_vectorized, variate_params, self.log, 
                         'gradient', 'finite_difference', 
                         {'stepsize': grad_stepsize})

        # Optimize
        vector_optimizer = get_optimizer(backend_obj_vectorized, variate_params, 
                                         optimizer_dict = {'method': 'vgd', 
                                                           'tol': 10**(-9), 
                                                           'jac': jac, 
                                                           'maxiter': niter, 
                                                           'optimizer_options' : 
                                                           {'stepsize': stepsize}, 
                                                           'save_intermediate': True
                                                          })
        vector_optimizer()

    def test_scipy_optimizers_global(self):
        " Check that final value of all scipy MINIMIZE_METHODS optimizers agrees with pre-computed optimized value."

        # Create problem instance, cost function, and gradient functions
        backend_obj_vectorized, variate_params = self.__backend_params(cost_hamiltonian_1, 4)        

        niter = 5
        stepsize = 0.001
        y_precomp = [-2.4345058914425626, -2.5889608823632795, -2.588960865651421, -2.5889608823632786, -2.5889608823632795, -2.588960882363273, -2.5889608823632786,
                     0.7484726235465329, -2.588960882363272, -2.588960882363281, -2.5889608823632786, -2.5889608823632795, -2.5889608823632786, -2.5889608823632786]
        optimizer_dicts = []
        for method in MINIMIZE_METHODS:
            optimizer_dicts.append(
                {'method': method, 'maxiter': niter, 'tol': 10**(-9)})

        for i, optimizer_dict in enumerate(optimizer_dicts):

            optimizer_dict['jac'] = derivative(backend_obj_vectorized, 
                variate_params, self.log, 'gradient', 'finite_difference')
            optimizer_dict['hess'] = derivative(backend_obj_vectorized, 
                variate_params, self.log, 'hessian', 'finite_difference')

            # Optimize
            vector_optimizer = get_optimizer(
                backend_obj_vectorized, variate_params, optimizer_dict=optimizer_dict)
            vector_optimizer()

            y_opt = vector_optimizer.qaoa_result.intermediate['cost']

            assert np.isclose(y_precomp[i], y_opt[-1], rtol=1e-04,
                              atol=1e-04), f"{optimizer_dict['method']} failed the test."

    def test_gradient_optimizers_global(self):
        " Check that final value of all implemented gradient optimizers agrees with pre-computed optimized value."

        backend_obj_vectorized, variate_params = self.__backend_params(cost_hamiltonian_1, 4)        
        niter = 10
        stepsize = 0.001

        # pre-computed final optimized costs
        y_precomp = [-2.4212581335011456, -2.4246393953483825, -
                     2.47312715451289, -2.5031221706241906]

        optimizer_dicts = []
        optimizer_dicts.append({'method': 'vgd', 'tol': 10**(-9), 'jac': 'finite_difference', 'maxiter': niter,
                                'optimizer_options' : {'stepsize': stepsize}})
        optimizer_dicts.append({'method': 'newton', 'tol': 10**(-9), 'jac': 'finite_difference', 'hess': 'finite_difference', 'maxiter': niter, 
                                'optimizer_options' : {'stepsize': stepsize}})
        optimizer_dicts.append({'method': 'natural_grad_descent', 'tol': 10**(-9), 'jac': 'finite_difference', 'maxiter': niter, 
                                'optimizer_options' : {'stepsize': 0.01}})
        optimizer_dicts.append({'method': 'rmsprop', 'tol': 10**(-9), 'jac': 'finite_difference', 'maxiter': niter,
                                'optimizer_options' : {'stepsize': stepsize, 'decay': 0.9, 'eps': 1e-07}})

        for i, optimizer_dict in enumerate(optimizer_dicts):

            # Optimize
            vector_optimizer = get_optimizer(backend_obj_vectorized, variate_params, optimizer_dict=optimizer_dict)
            vector_optimizer()

            y_opt = vector_optimizer.qaoa_result.intermediate['cost']

            assert np.isclose(y_precomp[i], y_opt[-1], rtol=1e-04, atol=1e-04), f"{optimizer_dict['method']} method failed the test."

    def test_gradient_optimizers_cans(self):
        n_qubits = 10
        graph = nw.circulant_graph(n_qubits, [1])
        cost_hamil = MinimumVertexCover(graph, field=1, penalty=10).qubo.hamiltonian
        backend_obj_vectorized, variate_params = self.__backend_params(cost_hamil, n_qubits)

        optimizer_list = [  {   'method':method, 'jac':'param_shift', 'maxiter':1000, 
                                'optimizer_options':{'stepsize': 0.0001,'n_shots_min':5, 'n_shots_max':200, 'n_shots_budget': 1000}  }    
                            for method in ['cans', 'icans']     ]

        for optimizer_dict in optimizer_list:
            # optimize qaoa
            vector_optimizer = get_optimizer(backend_obj_vectorized, variate_params, optimizer_dict=optimizer_dict)
            vector_optimizer()

            # check that the final cost is the optimal value
            assert int(vector_optimizer.qaoa_result.most_probable_states['bitstring_energy']) == 5, '{} failed the test.'.format(optimizer_dict['method'])
            # check that the number of shots are correct
            n_shots_used = np.array(vector_optimizer.qaoa_result.n_shots)
            assert np.sum(np.int0(n_shots_used<5) + np.int8(n_shots_used>200))==0 , 'Optimizer {} did not use the correct number of shots.'.format(optimizer_dict['method'])
            # check n_shots_budget worked
            assert  np.sum(20*n_shots_used[:-1].T[0] + 40*n_shots_used[:-1].T[1]) < 1000 and np.sum(20*n_shots_used.T[0] + 40*n_shots_used.T[1]) >= 1000, \
                    'Optimizer {} did not use the correct number of shots.'.format(optimizer_dict['method'])

            assert np.sum(n_shots_used<5) == 0 , 'Optimizer {} did not use the correct number of shots.'.format(optimizer_dict['method'])


    def test_gradient_descent_step(self):
        '''
        Check that implemented gradient descent takes the first two steps correctly.
        '''
        
        backend_obj_vectorized, variate_params = self.__backend_params(cost_hamiltonian_1, 4)        
        niter = 5
        grad_stepsize = 0.0001
        stepsize = 0.001

        params_array = variate_params.raw().copy()
        jac = derivative(backend_obj_vectorized, variate_params, self.log, 
                         'gradient', 'finite_difference', 
                         {'stepsize': grad_stepsize})

        # Optimize
        vector_optimizer = get_optimizer(backend_obj_vectorized, variate_params, optimizer_dict={
                                         'method': 'vgd', 'tol': 10**(-9), 'jac': jac, 'maxiter': niter,
                                         'optimizer_options' : {'stepsize': stepsize}})
        vector_optimizer()
        y_opt = vector_optimizer.qaoa_result.intermediate['cost'][1:4]

        # Stepwise optimize
        def step(x0):
            x1 = x0 - stepsize*jac(x0)

            variate_params.update_from_raw(x1)

            return [x1, np.real(backend_obj_vectorized.expectation(variate_params))]

        x0 = params_array
        variate_params.update_from_raw(x0)
        y0 = backend_obj_vectorized.expectation(variate_params)
        [x1, y1] = step(x0)
        [x2, y2] = step(x1)

        y = [y0, y1, y2]

        for i, yi in enumerate(y):
            assert np.isclose(yi, y_opt[i], rtol=1e-05, atol=1e-05)

    def test_newton_step(self):
        '''
        Check that implemented Newton descent takes the first two steps correctly.
        '''
        
        backend_obj_vectorized, variate_params = self.__backend_params(cost_hamiltonian_1, 4)        
        niter = 5
        grad_stepsize = 0.0001
        stepsize = 0.001

        params_array = variate_params.raw().copy()
        jac = derivative(backend_obj_vectorized, variate_params, self.log, 
                         'gradient', 'finite_difference', 
                         {'stepsize': grad_stepsize})
        hess = derivative(backend_obj_vectorized, variate_params, self.log, 
                          'hessian', 'finite_difference', 
                          {'stepsize': grad_stepsize})

        # Optimize
        vector_optimizer = get_optimizer(backend_obj_vectorized, variate_params, optimizer_dict={
                                         'method': 'newton', 'tol': 10**(-9), 'jac': jac, 'hess': hess, 'maxiter': niter,
                                         'optimizer_options' : {'stepsize': stepsize}})
        vector_optimizer()
        y_opt = vector_optimizer.qaoa_result.intermediate['cost'][1:4]

        # Stepwise optimize
        def step(x0):
            scaled_gradient = np.linalg.solve(hess(x0), jac(x0))
            x1 = x0 - stepsize*scaled_gradient
            variate_params.update_from_raw(x1)
            return [x1, np.real(backend_obj_vectorized.expectation(variate_params))]

        x0 = params_array
        variate_params.update_from_raw(x0)
        y0 = backend_obj_vectorized.expectation(variate_params)
        [x1, y1] = step(x0)
        [x2, y2] = step(x1)

        y = [y0, y1, y2]

        for i, yi in enumerate(y):
            assert np.isclose(yi, y_opt[i], rtol=1e-05, atol=1e-05)

    def test_natural_gradient_descent_step(self):
        '''
        Check that implemented natural gradient descent takes the first two steps correctly.
        '''

        backend_obj_vectorized, variate_params = self.__backend_params(cost_hamiltonian_1, 4)        
        niter = 5
        grad_stepsize = 0.0001
        stepsize = 0.001

        params_array = variate_params.raw().copy()
        jac = derivative(backend_obj_vectorized, variate_params, self.log, 
                         'gradient', 'finite_difference', 
                         {'stepsize': grad_stepsize})

        # Optimize
        vector_optimizer = get_optimizer(backend_obj_vectorized, variate_params, optimizer_dict={
                                         'method': 'natural_grad_descent', 'tol': 10**(-9), 'jac': jac, 'maxiter': niter,
                                         'optimizer_options' : {'stepsize': stepsize}})
        vector_optimizer()
        y_opt = vector_optimizer.qaoa_result.intermediate['cost'][1:4]

        # Stepwise optimize
        def step(x0):
            qfim_ = qfim(backend_obj_vectorized, variate_params, self.log)
            scaled_gradient = np.linalg.solve(qfim_(x0), jac(x0))
            x1 = x0 - stepsize*scaled_gradient
            variate_params.update_from_raw(x1)
            return [x1, np.real(backend_obj_vectorized.expectation(variate_params))]

        x0 = params_array
        variate_params.update_from_raw(x0)
        y0 = backend_obj_vectorized.expectation(variate_params)
        [x1, y1] = step(x0)
        [x2, y2] = step(x1)

        y = [y0, y1, y2]

        for i, yi in enumerate(y):
            assert np.isclose(yi, y_opt[i], rtol=1e-05, atol=1e-05)

    def test_rmsprop_step(self):
        '''
        Check that implemented RMSProp takes the first two steps correctly.
        '''
        
        backend_obj_vectorized, variate_params = self.__backend_params(cost_hamiltonian_1, 4)    
        niter = 5
        grad_stepsize = 0.0001
        stepsize = 0.001

        params_array = variate_params.raw().copy()
        jac = derivative(backend_obj_vectorized, variate_params, self.log, 
                         'gradient', 'finite_difference', 
                         {'stepsize': grad_stepsize})

        decay = 0.9
        eps = 1e-07

        # Optimize
        vector_optimizer = get_optimizer(backend_obj_vectorized, variate_params, optimizer_dict={
                                         'method': 'rmsprop', 'tol': 10**(-9), 'jac': jac, 'maxiter': niter,
                                         'optimizer_options' : {'stepsize': stepsize, 'decay': decay, 'eps': eps}})
                                         
        vector_optimizer()
        y_opt = vector_optimizer.qaoa_result.intermediate['cost'][1:4]

        # Stepwise optimize
        def step(x0, sqgrad0):
            sqgrad = decay*sqgrad0 + (1-decay)*jac(x0)**2
            x1 = x0 - stepsize*jac(x0)/(np.sqrt(sqgrad) + eps)
            variate_params.update_from_raw(x1)
            return [x1, np.real(backend_obj_vectorized.expectation(variate_params)), sqgrad0]

        x0 = params_array
        variate_params.update_from_raw(x0)
        y0 = backend_obj_vectorized.expectation(variate_params)
        sqgrad0 = jac(x0)**2
        [x1, y1, sqgrad1] = step(x0, sqgrad0)
        [x2, y2, sqgrad2] = step(x1, sqgrad1)

        y = [y0, y1, y2]

        for i, yi in enumerate(y):
            assert np.isclose(yi, y_opt[i], rtol=1e-05, atol=1e-05)
            
    def test_optimize_loop_crash(self):
        
        """
        This tests that the optimization loop doesnt crash silently.
        An Exception gets raised.
        """
        
        cost_hamil = Hamiltonian([PauliOp('ZZ', (0, 1)), PauliOp('ZZ', (1, 2)), PauliOp(
            'ZZ', (0, 3)), PauliOp('Z', (2,)), PauliOp('Z', (1,))], [1, 1.1, 1.5, 2, -0.8], 0.8)
        mixer_hamil = X_mixer_hamiltonian(n_qubits=4)
        qaoa_descriptor = QAOADescriptor(cost_hamil, mixer_hamil, p=2)
        device = create_device('local','vectorized')
        backend_obj_vectorized = get_qaoa_backend(qaoa_descriptor,device)
        variate_params = create_qaoa_variational_params(
            qaoa_descriptor, 'standard', 'ramp')
        niter = 5

        # Optimize
        vector_optimizer = get_optimizer(backend_obj_vectorized, variate_params, 
                                         optimizer_dict = {'method': 'nelder-mead',
                                                           'maxiter': niter,
                                                          })
        vector_optimizer.vqa.expectation = Mock(side_effect = Exception("Error!"))
        self.assertRaises(Exception, lambda: vector_optimizer.optimize())
        
        # Check that QAOA Result exists
        self.assertEqual(type(vector_optimizer.qaoa_result), QAOAResult)
    
    @classmethod
    def tearDownClass(cls):
        
        output_csv = ['oq_saved_info_job_ids.csv', 'oq_saved_info_param_log.csv']
        for each_csv in output_csv:
            if (os.path.exists(each_csv)):
                os.remove(each_csv)
            else:
                raise FileNotFoundError('Unable to remove the generated csv file: {}'.format(each_csv))
        

if __name__ == "__main__":
    with warnings.catch_warnings():
        warnings.simplefilter('ignore', category=PendingDeprecationWarning)
        unittest.main()



from .knapsack import Knapsack, SlackFreeKnapsack
from .maximumcut import MaximumCut
from .minimumvertexcover import MinimumVertexCover
from .numberpartition import NumberPartition
from .shortestpath import ShortestPath
from .tsp import TSP
from .converters import FromDocplex2IsingModel
from .qubo import QUBO



import unittest
import json
import os
import itertools
import subprocess
import pytest

from openqaoa.backends import DeviceLocal
from openqaoa.backends.devices_core import SUPPORTED_LOCAL_SIMULATORS
from openqaoa_qiskit.backends import DeviceQiskit
from openqaoa_braket.backends import DeviceAWS
from openqaoa_azure.backends import DeviceAzure


class TestingDeviceQiskit(unittest.TestCase):

    """These tests check the Object used to access IBMQ and their available
    QPUs can be established.
    For any tests using provided credentials, the tests will only pass if those
    details provided are correct/valid with IBMQ.

    Please ensure that the provided api token, hub, group and project in the 
    crendentials.json are correct.
    All of these can be found in your IBMQ Account Page.
    """
    @pytest.mark.api
    def setUp(self):

        self.HUB = 'ibm-q'
        self.GROUP = 'open'
        self.PROJECT = 'main'
    
    @pytest.mark.api
    def test_changing_provider(self):
        
        """
        This test checks that the specified hub,group and project in the 
        initialisation of DeviceQiskit changes the provider to the appropriate
        destination.
        """
        
        device_obj = DeviceQiskit(device_name='ibmq_manila')
        device_obj.check_connection()
        
        self.assertEqual(device_obj.provider.credentials.hub, self.HUB)
        self.assertEqual(device_obj.provider.credentials.group, self.GROUP)
        self.assertEqual(device_obj.provider.credentials.project, self.PROJECT)
        
        device_obj2 = DeviceQiskit(device_name='ibmq_manila', 
                                   hub='ibm-q-startup')
        device_obj2.check_connection()
        self.assertEqual(device_obj2.provider.credentials.hub, 'ibm-q-startup')
    
    @pytest.mark.api
    def test_check_connection_provider_no_backend_wrong_hub_group_project(self):
        
        """
        If the wrong hub, group or project is specified, check_connection should 
        return False.
        The provider_connected attribute should be updated to False.
        Since the API Token is loaded from save_account, the api token will be
        checked by Qiskit.
        """
        
        for each_combi in itertools.product(['invalid_hub', None], 
                                            ['invalid_group', None], 
                                            ['invalid_project', None]):
            
            if each_combi != (None, None, None):
                
                device_obj = DeviceQiskit(device_name='',
                                          hub=each_combi[0], 
                                          group=each_combi[1],
                                          project=each_combi[2])
            
                self.assertEqual(device_obj.check_connection(), False)
                self.assertEqual(device_obj.provider_connected, False)
                self.assertEqual(device_obj.qpu_connected, None)

    @pytest.mark.api
    def test_check_connection_provider_no_backend_provided_credentials(self):
        
        """
        If no information about the device name, but the credentials
        used are correct, check_connection should return True.
        The provider_connected attribute should be updated to True.
        """

        device_obj = DeviceQiskit(device_name='',
                                  hub=self.HUB, group=self.GROUP,
                                  project=self.PROJECT)

        self.assertEqual(device_obj.check_connection(), True)
        self.assertEqual(device_obj.provider_connected, True)
        self.assertEqual(device_obj.qpu_connected, None)

    @pytest.mark.api
    def test_check_connection_provider_right_backend_provided_credentials(self):
        
        """
        If the correct device name is provided and the credentials
        used are correct, check_connection should return True.
        The provider_connected attribute should be updated to True.
        The qpu_connected attribute should be updated to True.
        """

        device_obj = DeviceQiskit(device_name='', 
                                  hub=self.HUB, group=self.GROUP,
                                  project=self.PROJECT)

        device_obj.check_connection()
        valid_qpu_name = device_obj.available_qpus[0]

        device_obj = DeviceQiskit(device_name=valid_qpu_name, 
                                  hub=self.HUB, group=self.GROUP,
                                  project=self.PROJECT)

        self.assertEqual(device_obj.check_connection(), True)
        self.assertEqual(device_obj.provider_connected, True)
        self.assertEqual(device_obj.qpu_connected, True)

    @pytest.mark.api
    def test_check_connection_provider_wrong_backend_provided_credentials(self):
        
        """
        If device name provided is incorrect, and not empty, and the credentials
        used are correct, check_connection should return False.
        The provider_connected attribute should be updated to True.
        The qpu_connected attribute should be updated to False.
        """

        device_obj = DeviceQiskit(device_name='random_invalid_backend', 
                                  hub=self.HUB, group=self.GROUP,
                                  project=self.PROJECT)

        self.assertEqual(device_obj.check_connection(), False)
        self.assertEqual(device_obj.provider_connected, True)
        self.assertEqual(device_obj.qpu_connected, False)


class TestingDeviceLocal(unittest.TestCase):
    
    """
    This tests check that the Device Object created for local devices have the
    appropriate behaviour.
    """
    
    def test_supported_device_names(self):
        
        for each_device_name in SUPPORTED_LOCAL_SIMULATORS:
            device_obj = DeviceLocal(each_device_name)
            
            self.assertEqual(device_obj.check_connection(), True)
    
    def test_unsupported_device_names(self):
        
        device_obj = DeviceLocal('unsupported_device')
        
        self.assertEqual(device_obj.check_connection(), False)


class TestingDeviceAWS(unittest.TestCase):
    
    """These tests check the Object used to access AWS Braket and their 
    available QPUs can be established.

    For any tests using provided credentials, the tests will only pass if those
    details provided are correct/valid with AWS Braket.
    """
    
    @pytest.mark.api
    def test_changing_aws_region(self):
        
        device_obj = DeviceAWS(device_name='arn:aws:braket:::device/quantum-simulator/amazon/sv1')
        
        device_obj.check_connection()
        default_region = device_obj.aws_region
        
        self.assertEqual('us-east-1', default_region)
        
        device_obj = DeviceAWS(device_name='arn:aws:braket:::device/quantum-simulator/amazon/sv1', aws_region='us-west-1')
        
        device_obj.check_connection()
        custom_region = device_obj.aws_region
        
        self.assertEqual('us-west-1', custom_region)
        
    @pytest.mark.api
    def test_changing_s3_bucket_names(self):
        
        device_obj = DeviceAWS(device_name='arn:aws:braket:::device/quantum-simulator/amazon/sv1', s3_bucket_name='random_new_name')
        
        device_obj.check_connection()
        custom_bucket = device_obj.s3_bucket_name
        
        self.assertEqual('random_new_name', custom_bucket)
        
    @pytest.mark.api      
    def test_check_connection_provider_no_backend_provided_credentials(self):
        
        """
        If no information about the device name, but the credentials
        used are correct, check_connection should return True.
        The provider_connected attribute should be updated to True.
        """

        device_obj = DeviceAWS(device_name='')

        self.assertEqual(device_obj.check_connection(), True)
        self.assertEqual(device_obj.provider_connected, True)
        self.assertEqual(device_obj.qpu_connected, None)


    @pytest.mark.api
    def test_check_connection_provider_right_backend_provided_credentials(self):
        
        """
        If the correct device name is provided and the credentials
        used are correct, check_connection should return True.
        The provider_connected attribute should be updated to True.
        The qpu_connected attribute should be updated to True.
        """

        device_obj = DeviceAWS(device_name='')

        device_obj.check_connection()
        valid_qpu_name = device_obj.available_qpus[0]

        device_obj = DeviceAWS(device_name=valid_qpu_name)

        self.assertEqual(device_obj.check_connection(), True)
        self.assertEqual(device_obj.provider_connected, True)
        self.assertEqual(device_obj.qpu_connected, True)


    @pytest.mark.api
    def test_check_connection_provider_wrong_backend_provided_credentials(self):
        
        """
        If device name provided is incorrect, and not empty, and the credentials
        used are correct, check_connection should return False.
        The provider_connected attribute should be updated to True.
        The qpu_connected attribute should be updated to False.
        """

        device_obj = DeviceAWS(device_name='random_invalid_backend')

        self.assertEqual(device_obj.check_connection(), False)
        self.assertEqual(device_obj.provider_connected, True)
        self.assertEqual(device_obj.qpu_connected, False)
        

class TestingDeviceAzure(unittest.TestCase):
    
    """These tests check the Object used to access Azure and their 
    available QPUs can be established.

    For any tests using provided credentials, the tests will only pass if those
    details provided are correct/valid with Azure.
    """
    
    @pytest.mark.api
    def setUp(self):
        
        bashCommand = "az resource list"
        process = subprocess.Popen(bashCommand.split(), stdout=subprocess.PIPE)
        output, error = process.communicate()
        
        if error is not None:
            print(error)
            raise Exception('You must have the Azure CLI installed and must be logged in to use the Azure Quantum Backends')
        else:
            output_json = json.loads(output)
            output_json_s = [each_json for each_json in output_json if each_json['name'] == 'TestingOpenQAOA'][0]
            self.RESOURCE_ID = output_json_s['id']
            self.AZ_LOCATION = output_json_s['location']
            
    @pytest.mark.api
    def test_check_connection_provider_no_resource_id(self):
        
        """
        If no information about about the workspace is provided, the resource id
        or az location, check_connection and provider_connected should return False.
        """
        
        for resource_id, az_location in itertools.product(['', self.RESOURCE_ID], 
                                                          ['', self.AZ_LOCATION]):
            
            if not (resource_id == self.RESOURCE_ID and az_location == self.AZ_LOCATION):
            
                device_obj = DeviceAzure(device_name='', 
                                         resource_id=resource_id,
                                         az_location=az_location)

                self.assertEqual(device_obj.check_connection(), False)
                self.assertEqual(device_obj.provider_connected, False)
                self.assertEqual(device_obj.qpu_connected, None)
        
        
    @pytest.mark.api      
    def test_check_connection_provider_no_backend_provided_credentials(self):
        
        """
        If no information about the device name, but the credentials
        used are correct, check_connection should return True.
        The provider_connected attribute should be updated to True.
        """

        device_obj = DeviceAzure(device_name='', resource_id=self.RESOURCE_ID,
                                 az_location=self.AZ_LOCATION)

        self.assertEqual(device_obj.check_connection(), True)
        self.assertEqual(device_obj.provider_connected, True)
        self.assertEqual(device_obj.qpu_connected, None)


    @pytest.mark.api
    def test_check_connection_provider_right_backend_provided_credentials(self):
        
        """
        If the correct device name is provided and the credentials
        used are correct, check_connection should return True.
        The provider_connected attribute should be updated to True.
        The qpu_connected attribute should be updated to True.
        """

        device_obj = DeviceAzure(device_name='', resource_id=self.RESOURCE_ID,
                                 az_location=self.AZ_LOCATION)

        device_obj.check_connection()
        valid_qpu_name = device_obj.available_qpus[0]

        device_obj = DeviceAzure(device_name=valid_qpu_name, resource_id=self.RESOURCE_ID, 
                                 az_location=self.AZ_LOCATION)

        self.assertEqual(device_obj.check_connection(), True)
        self.assertEqual(device_obj.provider_connected, True)
        self.assertEqual(device_obj.qpu_connected, True)


    @pytest.mark.api
    def test_check_connection_provider_wrong_backend_provided_credentials(self):
        
        """
        If device name provided is incorrect, and not empty, and the credentials
        used are correct, check_connection should return False.
        The provider_connected attribute should be updated to True.
        The qpu_connected attribute should be updated to False.
        """

        device_obj = DeviceAzure(device_name='invalid_backend', resource_id=self.RESOURCE_ID,
                                 az_location=self.AZ_LOCATION)

        self.assertEqual(device_obj.check_connection(), False)
        self.assertEqual(device_obj.provider_connected, True)
        self.assertEqual(device_obj.qpu_connected, False)

if __name__ == '__main__':
    unittest.main()



import time

from .qaoa_result import QAOAResult
from ..workflow_properties import CircuitProperties
from ..baseworkflow import Workflow
from ...backends.devices_core import DeviceLocal
from ...backends.qaoa_backend import get_qaoa_backend
from ...problems import QUBO
from ...qaoa_components import (
    Hamiltonian,
    QAOADescriptor,
    create_qaoa_variational_params,
)
from ...utilities import get_mixer_hamiltonian
from ...optimizers.qaoa_optimizer import get_optimizer


class QAOA(Workflow):
    """
    A class implementing a QAOA workflow end to end.

    It's basic usage consists of
    1. Initialization
    2. Compilation
    3. Optimization

    .. note::
        The attributes of the QAOA class should be initialized using the set methods of QAOA.
        For example, to set the circuit's depth to 10 you should run `set_circuit_properties(p=10)`

    Attributes
    ----------
    device: `DeviceBase`
        Device to be used by the optimizer
    circuit_properties: `CircuitProperties`
        The circuit properties of the QAOA workflow. Use to set depth `p`,
        choice of parameterization, parameter initialisation strategies, mixer hamiltonians.
        For a complete list of its parameters and usage please see the method `set_circuit_properties`
    backend_properties: `BackendProperties`
        The backend properties of the QAOA workflow. Use to set the backend properties
        such as the number of shots and the cvar values.
        For a complete list of its parameters and usage please see the method `set_backend_properties`
    classical_optimizer: `ClassicalOptimizer`
        The classical optimiser properties of the QAOA workflow. Use to set the
        classical optimiser needed for the classical optimisation part of the QAOA routine.
        For a complete list of its parameters and usage please see the method `set_classical_optimizer`
    local_simulators: `list[str]`
        A list containing the available local simulators
    cloud_provider: `list[str]`
        A list containing the available cloud providers
    mixer_hamil: Hamiltonian
        The desired mixer hamiltonian
    cost_hamil: Hamiltonian
        The desired mixer hamiltonian
    qaoa_descriptor: QAOADescriptor
        the abstract and backend-agnostic representation of the underlying QAOA parameters
    variate_params: QAOAVariationalBaseParams
        The variational parameters. These are the parameters to be optimised by the classical optimiser
    backend: VQABaseBackend
        The openQAOA representation of the backend to be used to execute the quantum circuit
    optimizer: OptimizeVQA
        The classical optimiser
    result: `Result`
        Contains the logs of the optimisation process
    compiled: `Bool`
        A boolean flag to check whether the QAOA object has been correctly compiled at least once

    Examples
    --------
    Examples should be written in doctest format, and should illustrate how
    to use the function.

    >>> q = QAOA()
    >>> q.compile(QUBO)
    >>> q.optimise()

    Where `QUBO` is a an instance of `openqaoa.problems.problem.QUBO`

    If you want to use non-default parameters:

    >>> q_custom = QAOA()
    >>> q_custom.set_circuit_properties(
            p=10,
            param_type='extended',
            init_type='ramp',
            mixer_hamiltonian='x'
        )
    >>> q_custom.set_device_properties(
            device_location = 'qcs', device_name='Aspen-11',
            cloud_credentials = {
                'name' : "Aspen11", 'as_qvm':True,
                'execution_timeout' : 10, 'compiler_timeout':10
            }
        )
    >>> q_custom.set_backend_properties(n_shots=200, cvar_alpha=1)
    >>> q_custom.set_classical_optimizer(method='nelder-mead', maxiter=2)
    >>> q_custom.compile(qubo_problem)
    >>> q_custom.optimize()
    """

    results_class = QAOAResult

    def __init__(self, device=DeviceLocal("vectorized")):
        """
        Initialize the QAOA class.

        Parameters
        ----------
        device: `DeviceBase`
            Device to be used by the optimizer. Default is using the local 'vectorized' simulator.
        """
        super().__init__(device)
        self.circuit_properties = CircuitProperties()

        # change header algorithm to qaoa
        self.header["algorithm"] = "qaoa"

    def set_circuit_properties(self, **kwargs):
        """
        Specify the circuit properties to construct QAOA circuit

        Parameters
        ----------
        qubit_register: `list`
            Select the desired qubits to run the QAOA program. Meant to be used as a qubit
            selector for qubits on a QPU. Defaults to a list from 0 to n-1 (n = number of qubits)
        p: `int`
            Depth `p` of the QAOA circuit
        q: `int`
            Analogue of `p` of the QAOA circuit in the Fourier parameterization
        param_type: `str`
            Choose the QAOA circuit parameterization. Currently supported parameterizations include:
            `'standard'`: Standard QAOA parameterization
            `'standard_w_bias'`: Standard QAOA parameterization with a separate parameter for single-qubit terms.
            `'extended'`: Individual parameter for each qubit and each term in the Hamiltonian.
            `'fourier'`: Fourier circuit parameterization
            `'fourier_extended'`: Fourier circuit parameterization with individual parameter
            for each qubit and term in Hamiltonian.
            `'fourier_w_bias'`: Fourier circuit parameterization with a separate
            parameter for single-qubit terms
        init_type: `str`
            Initialisation strategy for the QAOA circuit parameters. Allowed init_types:
            `'rand'`: Randomly initialise circuit parameters
            `'ramp'`: Linear ramp from Hamiltonian initialisation of circuit
            parameters (inspired from Quantum Annealing)
            `'custom'`: User specified initial circuit parameters
        mixer_hamiltonian: `str`
            Parameterisation of the mixer hamiltonian:
            `'x'`: Randomly initialise circuit parameters
            `'xy'`: Linear ramp from Hamiltonian initialisation of circuit
        mixer_qubit_connectivity: `[Union[List[list],List[tuple], str]]`
            The connectivity of the qubits in the mixer Hamiltonian. Use only if
            `mixer_hamiltonian = xy`. The user can specify the connectivity as a list of lists,
            a list of tuples, or a string chosen from ['full', 'chain', 'star'].
        mixer_coeffs: `list`
            The coefficients of the mixer Hamiltonian. By default all set to -1
        annealing_time: `float`
            Total time to run the QAOA program in the Annealing parameterization (digitised annealing)
        linear_ramp_time: `float`
            The slope(rate) of linear ramp initialisation of QAOA parameters.
        variational_params_dict: `dict`
            Dictionary object specifying the initial value of each circuit parameter for
            the chosen parameterization, if the `init_type` is selected as `'custom'`.
            For example, for standard params set {'betas': [0.1, 0.2, 0.3], 'gammas': [0.1, 0.2, 0.3]}
        """

        for key, value in kwargs.items():
            if hasattr(self.circuit_properties, key):
                pass
            else:
                raise ValueError("Specified argument is not supported by the circuit")
        self.circuit_properties = CircuitProperties(**kwargs)

        return None

    def compile(self, problem: QUBO = None, verbose: bool = False):
        """
        Initialise the trainable parameters for QAOA according to the specified
        strategies and by passing the problem statement

        .. note::
            Compilation is necessary because it is the moment where the problem statement and
            the QAOA instructions are used to build the actual QAOA circuit.

        .. tip::
            Set Verbose to false if you are running batch computations!

        Parameters
        ----------
        problem: `Problem`
            QUBO problem to be solved by QAOA
        verbose: bool
            Set True to have a summary of QAOA to displayed after compilation
        """

        # we compile the method of the parent class to genereate the id and
        # check the problem is a QUBO object and save it
        super().compile(problem=problem)

        self.cost_hamil = Hamiltonian.classical_hamiltonian(
            terms=problem.terms, coeffs=problem.weights, constant=problem.constant
        )

        self.mixer_hamil = get_mixer_hamiltonian(
            n_qubits=self.cost_hamil.n_qubits,
            mixer_type=self.circuit_properties.mixer_hamiltonian,
            qubit_connectivity=self.circuit_properties.mixer_qubit_connectivity,
            coeffs=self.circuit_properties.mixer_coeffs,
        )

        self.qaoa_descriptor = QAOADescriptor(
            self.cost_hamil, self.mixer_hamil, p=self.circuit_properties.p
        )
        self.variate_params = create_qaoa_variational_params(
            qaoa_descriptor=self.qaoa_descriptor,
            params_type=self.circuit_properties.param_type,
            init_type=self.circuit_properties.init_type,
            variational_params_dict=self.circuit_properties.variational_params_dict,
            linear_ramp_time=self.circuit_properties.linear_ramp_time,
            q=self.circuit_properties.q,
            seed=self.circuit_properties.seed,
            total_annealing_time=self.circuit_properties.annealing_time,
        )

        self.backend = get_qaoa_backend(
            qaoa_descriptor=self.qaoa_descriptor,
            device=self.device,
            **self.backend_properties.__dict__,
        )

        self.optimizer = get_optimizer(
            vqa_object=self.backend,
            variational_params=self.variate_params,
            optimizer_dict=self.classical_optimizer.asdict(),
        )

        self.compiled = True

        if verbose:
            print("\t \033[1m ### Summary ###\033[0m")
            print(f"OpenQAOA has been compiled with the following properties")
            print(
                f"Solving QAOA with \033[1m {self.device.device_name} \033[0m on"
                f"\033[1m{self.device.device_location}\033[0m"
            )
            print(
                f"Using p={self.circuit_properties.p} with {self.circuit_properties.param_type}"
                f"parameters initialized as {self.circuit_properties.init_type}"
            )

            if hasattr(self.backend, "n_shots"):
                print(
                    f"OpenQAOA will optimize using \033[1m{self.classical_optimizer.method}"
                    f"\033[0m, with up to \033[1m{self.classical_optimizer.maxiter}"
                    f"\033[0m maximum iterations. Each iteration will contain"
                    f"\033[1m{self.backend_properties.n_shots} shots\033[0m"
                )
            else:
                print(
                    f"OpenQAOA will optimize using \033[1m{self.classical_optimizer.method}\033[0m,"
                    "with up to \033[1m{self.classical_optimizer.maxiter}\033[0m maximum iterations"
                )

        return None

    def optimize(self, verbose=False):
        """
        A method running the classical optimisation loop
        """

        if self.compiled == False:
            raise ValueError("Please compile the QAOA before optimizing it!")

        # timestamp for the start of the optimization
        self.header["execution_time_start"] = time.time_ns()

        self.optimizer.optimize()
        # TODO: result and qaoa_result will differ
        self.result = self.optimizer.qaoa_result

        # timestamp for the end of the optimization
        self.header["execution_time_end"] = time.time_ns()

        if verbose:
            print(f"optimization completed.")
        return

    def _serializable_dict(
        self, complex_to_string: bool = False, intermediate_mesurements: bool = True
    ):
        """
        Returns all values and attributes of the object that we want to return in
        `asdict` and `dump(s)` methods in a dictionary.

        Parameters
        ----------
        complex_to_string: bool
            If True, complex numbers are converted to strings.
            This is useful for JSON serialization.

        Returns
        -------
        serializable_dict: dict
            A dictionary containing all the values and attributes of the object
            that we want to return in `asdict` and `dump(s)` methods.
        intermediate_mesurements: bool
            If True, intermediate measurements are included in the dump.
            If False, intermediate measurements are not included in the dump.
            Default is True.
        """

        # we call the _serializable_dict method of the parent class,
        # specifying the keys to delete from the results dictionary
        serializable_dict = super()._serializable_dict(
            complex_to_string, intermediate_mesurements
        )

        # we add the keys of the QAOA object that we want to return
        serializable_dict["data"]["input_parameters"]["circuit_properties"] = dict(
            self.circuit_properties
        )

        # include parameters in the header metadata
        serializable_dict["header"]["metadata"]["param_type"] = serializable_dict[
            "data"
        ]["input_parameters"]["circuit_properties"]["param_type"]
        serializable_dict["header"]["metadata"]["init_type"] = serializable_dict[
            "data"
        ]["input_parameters"]["circuit_properties"]["init_type"]
        serializable_dict["header"]["metadata"]["p"] = serializable_dict["data"][
            "input_parameters"
        ]["circuit_properties"]["p"]

        if (
            serializable_dict["data"]["input_parameters"]["circuit_properties"]["q"]
            is not None
        ):
            serializable_dict["header"]["metadata"]["q"] = serializable_dict["data"][
                "input_parameters"
            ]["circuit_properties"]["q"]

        return serializable_dict



from collections import Counter
import numpy as np
from pyquil import Program, gates, quilbase

from .devices import DevicePyquil
from openqaoa.backends.basebackend import (
    QAOABaseBackendShotBased,
    QAOABaseBackendCloud,
    QAOABaseBackendParametric,
)
from openqaoa.qaoa_components import QAOADescriptor
from openqaoa.qaoa_components.variational_parameters.variational_baseparams import QAOAVariationalBaseParams
from openqaoa.qaoa_components.ansatz_constructor.gatemap import RZZGateMap, SWAPGateMap
from openqaoa.utilities import generate_uuid


def check_edge_connectivity(executable: Program, device: DevicePyquil):

    '''
    Check that the program does not contain 2-qubit terms that is not present
    in the QPU's topology (to prevent quilc from crashing).

    Parameters
    ----------
    executable: `Program`
        pyQuil executable program.
    device: `DevicePyquil`
        An object of the class ``DevicePyquil`` which contains information on
        pyQuil's `QuantumComputer` object, used to extract the selected QPU's topology.

    Returns
    -------
        None
    """
    '''

    qpu_graph = device.quantum_computer.qubit_topology()

    instrs = [instr for instr in executable if type(instr) == quilbase.Gate]
    pair_instrs = [
        list(instr.get_qubits()) for instr in instrs if len(instr.get_qubits()) == 2
    ]

    for term in pair_instrs:
        if len(term) == 2:

            assert (
                term in qpu_graph.edges()
            ), f"Term {term} is not an edge on the QPU graph of {device.device_name}."


class QAOAPyQuilQPUBackend(
    QAOABaseBackendParametric, QAOABaseBackendCloud, QAOABaseBackendShotBased
):
    """
    A QAOA backend object for real Rigetti QPUs

    Parameters
    ----------
    device: `DevicePyquil`
        The device object to access pyquil devices with credentials.
    qaoa_descriptor: `QAOADescriptor`
        An object of the class ``QAOADescriptor`` which contains information on
        circuit construction and depth of the circuit.
    n_shots: `int`
        The number of shots to be taken for each circuit.
    prepend_state: `pyquil.Program`
        The state prepended to the circuit.
    append_state: `pyquil.Program`
        The state appended to the circuit.
    init_hadamard: `bool`
        Whether to apply a Hadamard gate to the beginning of the QAOA part of the circuit.
    cvar_alpha: `float`
        Conditional Value-at-Risk (CVaR) – a measure that takes into account only the tail of the
        probability distribution arising from the circut's count dictionary. Must be between 0 and 1. Check
        https://arxiv.org/abs/1907.04769 for further details.
    active_reset:
        Whether to use the pyQuil's active reset scheme to reset qubits between shots.
    rewiring:
        Rewiring scheme to be used for Pyquil.
        Either PRAGMA INITIAL_REWIRING "NAIVE" or PRAGMA INITIAL_REWIRING "PARTIAL".
        If None, pyquil defaults according to:
        NAIVE: The qubits used in all instructions in the program satisfy the
        topological constraints of the device.
        PARTIAL: Otherwise.
    """

    def __init__(
        self,
        device: DevicePyquil,
        qaoa_descriptor: QAOADescriptor,
        n_shots: int,
        prepend_state: Program,
        append_state: Program,
        init_hadamard: bool,
        cvar_alpha: float,
        active_reset: bool = False,
        rewiring: str = "",
        qubit_layout: list = [],
    ):

        QAOABaseBackendShotBased.__init__(
            self,
            qaoa_descriptor,
            n_shots,
            prepend_state,
            append_state,
            init_hadamard,
            cvar_alpha,
        )
        QAOABaseBackendCloud.__init__(self, device)

        self.active_reset = active_reset
        self.rewiring = rewiring
        self.qureg = self.qaoa_descriptor.qureg

        # self.qureg_placeholders = QubitPlaceholder.register(self.n_qubits)
        self.qubit_layout = self.qureg if qubit_layout == [] else qubit_layout
        self.qubit_mapping = dict(zip(self.qureg, self.qubit_layout))

        if self.prepend_state:
            assert self.n_qubits >= len(prepend_state.get_qubits()), (
                "Cannot attach a bigger circuit " "to the QAOA routine"
            )

        if self.device.n_qubits < self.n_qubits:
            raise Exception(
                "There are lesser qubits on the device than the number of qubits required for the circuit."
            )

        self.parametric_circuit = self.parametric_qaoa_circuit
        native_prog = self.device.quantum_computer.compiler.quil_to_native_quil(
            self.parametric_circuit
        )
        self.prog_exe = self.device.quantum_computer.compiler.native_quil_to_executable(
            native_prog
        )

        # Check program connectivity against QPU connectivity
        # TODO: reconcile with PRAGMA PRESERVE

        # check_edge_connectivity(self.prog_exe, device)

    def qaoa_circuit(self, params: QAOAVariationalBaseParams) -> Program:
        """
        Injects angles into created executable parametric circuit.

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`

        Returns
        -------
        `pyquil.Program`
            A pyquil.Program (executable) object.
        """
        angles_list = np.array(
            self.obtain_angles_for_pauli_list(self.abstract_circuit, params),
            dtype=float,
        )
        angle_declarations = list(self.parametric_circuit.declarations.keys())
        angle_declarations.remove("ro")
        for i, param_name in enumerate(angle_declarations):
            self.prog_exe.write_memory(region_name=param_name, value=angles_list[i])

        return self.prog_exe

    @property
    def parametric_qaoa_circuit(self) -> Program:
        """
        Creates a parametric QAOA circuit (pyquil.Program object),
        given the qubit pairs, single qubits with biases,
        and a set of circuit angles. Note that this function does not actually run
        the circuit.

        Parameters
        ----------
        params: `QAOAVariationalBaseParams`

        Returns
        -------
        `pyquil.Program`
            A pyquil.Program object.
        """
        if self.active_reset:
            parametric_circuit = Program(gates.RESET())
        else:
            parametric_circuit = Program()

        if self.rewiring != None:
            if self.rewiring in [
                'PRAGMA INITIAL_REWIRING "NAIVE"',
                'PRAGMA INITIAL_REWIRING "PARTIAL"',
                "",
            ]:
                parametric_circuit += Program(self.rewiring)
            else:
                raise ValueError(
                    "Rewiring command not recognized. Please use "
                    'PRAGMA INITIAL_REWIRING "NAIVE"'
                    " or "
                    'PRAGMA INITIAL_REWIRING "PARTIAL"'
                    ""
                )

        # declare the read-out register
        ro = parametric_circuit.declare("ro", "BIT", self.n_qubits)

        if self.prepend_state:
            parametric_circuit += self.prepend_state

        # Initial state is all |+>
        if self.init_hadamard:
            for i in self.qureg:
                parametric_circuit += gates.RZ(np.pi, self.qubit_mapping[i])
                parametric_circuit += gates.RX(np.pi / 2, self.qubit_mapping[i])
                parametric_circuit += gates.RZ(np.pi / 2, self.qubit_mapping[i])
                parametric_circuit += gates.RX(-np.pi / 2, self.qubit_mapping[i])

        # create a list of gates in order of application on quantum circuit
        for each_gate in self.abstract_circuit:
            gate_label = "".join(str(label) for label in each_gate.pauli_label)
            angle_param = parametric_circuit.declare(f"pauli{gate_label}", "REAL", 1)
            each_gate.rotation_angle = angle_param
            if isinstance(each_gate, RZZGateMap) or isinstance(each_gate, SWAPGateMap):
                decomposition = each_gate.decomposition("standard2")
            else:
                decomposition = each_gate.decomposition("standard")

            # using the list above, construct the circuit
            for each_tuple in decomposition:
                gate = each_tuple[0]()
                qubits, rotation_angle = each_tuple[1]
                if isinstance(qubits, list):
                    new_qubits = [self.qubit_mapping[qubit] for qubit in qubits]
                else:
                    new_qubits = self.qubit_mapping[qubits]
                parametric_circuit = gate.apply_pyquil_gate(
                    new_qubits, rotation_angle, parametric_circuit
                )

        if self.append_state:
            parametric_circuit += self.append_state

        # Measurement instructions
        for i, qubit in enumerate(self.qureg):
            parametric_circuit += gates.MEASURE(self.qubit_mapping[qubit], ro[i])

        parametric_circuit.wrap_in_numshots_loop(self.n_shots)

        return parametric_circuit

    def get_counts(self, params: QAOAVariationalBaseParams, n_shots=None) -> dict:
        """
        Execute the circuit and obtain the counts.

        Parameters
        ----------
        params: QAOAVariationalBaseParams
            The QAOA parameters - an object of one of the parameter classes, containing
            variable parameters.
        n_shots: int
            The number of times to run the circuit. If None, n_shots is set to the default: self.n_shots

        Returns
        -------
        counts : dictionary
            A dictionary with the bitstring as the key and the number of counts as its value.
        """

        executable_program = self.qaoa_circuit(params)

        # if n_shots is given change the number of shots
        if n_shots is not None:
            executable_program.wrap_in_numshots_loop(n_shots)

        result = self.device.quantum_computer.run(executable_program)

        # we create an uuid for the job
        self.job_id = generate_uuid()

        # TODO: check the endian (big or little) ordering of measurement outcomes
        meas_list = [
            "".join(str(bit) for bit in bitstring)
            for bitstring in result.readout_data["ro"]
        ]

        # Expose counts
        counts = Counter(list(meas_list))
        self.measurement_outcomes = counts
        return counts

    def circuit_to_qasm(self, params: QAOAVariationalBaseParams) -> str:
        """
        A method to convert the pyQuil program to a OpenQASM string.
        """
        raise NotImplementedError()
        # qasm_program = self.device.quantum_computer.compiler.quil_to_qasm(self.qaoa_circuit(params))
        # return qasm_program

    def reset_circuit(self):
        """
        Reset self.program after performing a computation. Also handle active reset and rewirings.
        """
        pass



import unittest
import networkx as nx
import numpy as np
from random import randint, random
from openqaoa.problems import (
    NumberPartition, QUBO, TSP, Knapsack, ShortestPath,
    SlackFreeKnapsack, MaximumCut, MinimumVertexCover
)
from openqaoa.utilities import convert2serialize
from openqaoa.problems.helper_functions import create_problem_from_dict

def terms_list_equality(terms_list1, terms_list2):
    """
    Check the terms equality between two terms list
    where the order of edges do not matter.
    """
    if len(terms_list1) != len(terms_list2):
        bool = False
    else:
        for term1, term2 in zip(terms_list1, terms_list2):
            bool = True if (term1 == term2 or term1 == term2[::-1]) else False

    return bool


class TestProblem(unittest.TestCase):

    # TESTING QUBO CLASS METHODS
    def test_qubo_terms_and_weight_same_size(self):
        """
        Test that creating a QUBO problem with invalid terms and weights
        sizes raises an exception and check the constant is detected correctly
        """
        n = 2
        terms_wrong = [[0], [1]]
        terms_correct = [[0], [1], []]
        weights = [1, 2, 3]

        with self.assertRaises(ValueError):
            qubo_problem = QUBO(n, terms_wrong, weights)

        qubo_problem = QUBO(n, terms_correct, weights)
        self.assertEqual(qubo_problem.constant, 3)

    def test_qubo_cleaning_terms(self):
        """Test that cleaning terms works for a QUBO problem"""
        terms = [[1, 2], [0], [2, 3], [2, 1], [0]]
        weights = [3, 4, -3, -2, -1]

        cleaned_terms = [[1, 2], [0], [2, 3]]
        self.assertEqual(QUBO.clean_terms_and_weights(
            terms, weights)[0], cleaned_terms)

    def test_qubo_cleaning_weights(self):
        """Test that cleaning weights works for a QUBO problem"""
        terms = [[1, 2], [0], [2, 3], [2, 1], [0]]
        weights = [3, 4, -3, -2, -1]

        cleaned_weights = [1, 3, -3]
        self.assertEqual(QUBO.clean_terms_and_weights(
            terms, weights)[1], cleaned_weights)

    def test_qubo_ising_conversion(self):
        """Test that conversion to Ising formulation works for a QUBO problem"""
        # Small instance
        n = 2
        terms = [[1, 1], [0, 0]]
        weights = [3, 4]

        expected_ising_terms = [[0], [1], []]
        expected_ising_weights = [-2, -1.5, 3.5]

        ising_terms, ising_weights = QUBO.convert_qubo_to_ising(
            n, terms, weights)
        self.assertEqual(expected_ising_terms, ising_terms)
        self.assertEqual(expected_ising_weights, ising_weights)

        # Larger instance
        n = 4
        terms = [[1, 2], [0], [2, 3], [2, 1], [0]]
        weights = [3, 4, -3, -2, -1]

        expected_ising_terms = [[1, 2], [2, 3], [2, 1], [0], [1], [2], [3], []]
        expected_ising_weights = [
            0.75, -0.75, -0.5, -1.5, -0.25, 0.5, 0.75, 1.0]

        ising_terms, ising_weights = QUBO.convert_qubo_to_ising(
            n, terms, weights)
        self.assertEqual(expected_ising_terms, ising_terms)
        self.assertEqual(expected_ising_weights, ising_weights)

    def test_qubo_type_checking(self):
        """
        Checks if the type-checking returns the right error.
        """

        # n type-check
        n_list = [1.5, 'test', [], (), {}, np.array(1)]
        terms = [[0], []]
        weights = [1, 2]

        with self.assertRaises(TypeError) as e:
            for each_n in n_list:
                QUBO(each_n, terms, weights)
            self.assertEqual("The input parameter, n, has to be of type int",
                             str(e.exception))

        n_list = [-1, 0]
        with self.assertRaises(TypeError) as e:
            for each_n in n_list:
                QUBO(each_n, terms, weights)
            self.assertEqual("The input parameter, n, must be a positive integer greater than 0",
                             str(e.exception))

        # weights type-check
        n = 1
        terms = [[0], []]
        weights_list = [{'test': 'oh', 'test1': 'oh'}, np.array([1, 2])]

        for each_weights in weights_list:
            with self.assertRaises(TypeError) as e:
                QUBO(n, terms, each_weights)
            self.assertEqual("The input parameter weights must be of type of list or tuple",
                             str(e.exception))

        weights_list = [['test', 'oh'], [np.array(1), np.array(2)]]
        for each_weights in weights_list:
            with self.assertRaises(TypeError) as e:
                QUBO(n, terms, each_weights)
            self.assertEqual("The elements in weights list must be of type float or int.",
                             str(e.exception))

        # terms type-check
        n = 1
        terms_list = [{'test': [0], 'test1': []}]
        weights = [1, 2]
        for each_terms in terms_list:
            with self.assertRaises(TypeError) as e:
                QUBO(n, each_terms, weights)
            self.assertEqual("The input parameter terms must be of type of list or tuple",
                             str(e.exception))

    def test_qubo_metadata(self):
        """Test that metadata is correctly stored"""
        qubo_problem = QUBO.random_instance(3)
        qubo_problem.set_metadata({'tag1': 'value1', 'tag2': 'value2'})
        qubo_problem.set_metadata({'tag2': 'value2.0'})

        assert qubo_problem.metadata['tag1'] == 'value1', "qubo metadata is not well set"
        assert qubo_problem.metadata['tag2'] == 'value2.0', "qubo metadata is not well set, should have overwritten previous value"

        error = False
        try:
            qubo_problem.set_metadata({'tag10': complex(1, 2)})
        except:
            error = True
        assert error, "Should have thrown an error when setting metadata that is not json serializable"

        error = False
        try:
            qubo_problem.set_metadata({(1,2): 'value'})
        except:
            error = True
        assert error, "Should have thrown an error when setting key metadata that is not json serializable"

    def test_qubo_problem_instance_serializable(self):
        """ test that when problem instance is not serializable, it throws an error """
        
        qubo = QUBO.random_instance(3)

        error = False
        try:
            qubo.problem_instance={'tag10': complex(1, 2)}
        except:
            error = True
        assert error, "Should have thrown an error when setting qubo problem instance that is not json serializable"

        error = False
        try:
            qubo.problem_instance={(1,2): 'value'}
        except:
            error = True
        assert error, "Should have thrown an error when setting key qubo problem instance that is not json serializable"


    # TESTING NUMBER PARITION CLASS
    
    def test_number_partitioning_terms_weights_constant(self):
        """Test that Number Partitioning creates the correct terms, weights, constant"""
        list_numbers = [1, 2, 3]
        expected_terms = [[0, 1], [0, 2], [1, 2]]
        expected_weights = [4, 6, 12]
        expected_constant = 14

        np_problem = NumberPartition(list_numbers)
        qubo_problem = np_problem.qubo

        self.assertTrue(terms_list_equality(
            qubo_problem.terms, expected_terms))
        self.assertEqual(qubo_problem.weights, expected_weights)
        self.assertEqual(qubo_problem.constant, expected_constant)

    def test_number_partitioning_random_problem(self):
        """Test randomly generated NumberPartition problem"""
        # regenerate the same numbers randomly
        rng = np.random.default_rng(1234)
        random_numbers_list = list(map(int, rng.integers(1, 10, size=5)))
        manual_np_prob = NumberPartition(
            random_numbers_list).qubo

        np_prob_random = NumberPartition.random_instance(
            n_numbers=5, seed=1234).qubo

        self.assertTrue(terms_list_equality(
            np_prob_random.terms, manual_np_prob.terms))
        self.assertEqual(np_prob_random.weights, manual_np_prob.weights)
        self.assertEqual(np_prob_random.constant, manual_np_prob.constant)

    def test_num_part_type_checking(self):
        """
        Checks if the type-checking returns the right error.
        """

        # numbers type-check
        numbers_list = [(1, 2), {'test': 1, 'test1': 2}, np.array([1, 2])]

        for each_number in numbers_list:
            with self.assertRaises(TypeError) as e:
                NumberPartition(numbers=each_number)
            self.assertEqual("The input parameter, numbers, has to be a list",
                             str(e.exception))

        numbers_list = [[0.1, 1], [np.array(1), np.array(2)]]

        for each_number in numbers_list:
            with self.assertRaises(TypeError) as e:
                NumberPartition(numbers=each_number)
            self.assertEqual("The elements in numbers list must be of type int.",
                             str(e.exception))

    # TESTING MAXIMUMCUT CLASS

    def test_maximumcut_terms_weights_constant(self):
        """Test that MaximumCut creates a correct QUBO from the provided graph"""

        gr = nx.generators.random_graphs.fast_gnp_random_graph(n=10, p=0.8)
        gr_edges = [list(edge) for edge in gr.edges()]
        gr_weights = [1]*len(gr_edges)

        maxcut_prob_qubo = MaximumCut(gr).qubo

        self.assertTrue(terms_list_equality(gr_edges, maxcut_prob_qubo.terms))
        self.assertEqual(gr_weights, maxcut_prob_qubo.weights)
        self.assertEqual(0, maxcut_prob_qubo.constant)

    def test_maximumcut_random_problem(self):
        """Test MaximumCut random instance method"""

        seed = 1234
        gr = nx.generators.random_graphs.fast_gnp_random_graph(
            n=10, p=0.8, seed=seed)
        maxcut_manual_prob = MaximumCut(gr).qubo

        np.random.seed(1234)
        maxcut_random_prob = MaximumCut.random_instance(
            n_nodes=10, edge_probability=0.8, seed=seed).qubo

        self.assertTrue(terms_list_equality(
            maxcut_manual_prob.terms, maxcut_random_prob.terms))
        self.assertEqual(maxcut_manual_prob.weights,
                         maxcut_random_prob.weights)
        self.assertEqual(maxcut_manual_prob.constant,
                         maxcut_random_prob.constant)

    def test_maximumcut_type_checking(self):
        """
        Checks if the type-checking returns the right error.
        """

        # graph type-check
        graph_list = [(1, 2), {'node1': 1, 'node2': 2}, np.array([1, 2])]

        for each_graph in graph_list:
            with self.assertRaises(TypeError) as e:
                MaximumCut(G=each_graph)
            self.assertEqual("Input problem graph must be a networkx Graph.",
                             str(e.exception))

    # TESTING KNAPSACK CLASS

    def test_knapsack_terms_weights_constant(self):
        """Test that Knapsack creates the correct QUBO problem"""

        values = [2, 4, 3, 5]
        weights = [3, 6, 9, 1]
        weight_capacity = 15
        n_qubits = len(values) + int(np.ceil(np.log2(weight_capacity)))
        penalty = 2*max(values)
        knap_terms = [[0, 1], [0, 2], [0, 3], [1, 2], [1, 3], [2, 3], [4, 5], [4, 6], [4, 7], [5, 6],
                      [5, 7], [6, 7], [0, 4], [0, 5], [0, 6], [
                          0, 7], [1, 4], [1, 5], [1, 6], [1, 7],
                      [2, 4], [2, 5], [2, 6], [2, 7], [3, 4], [
                          3, 5], [3, 6], [3, 7], [0], [1], [2],
                      [3], [4], [5], [6], [7]]
        knap_weights = [10.0, 20.0, 40.0, 40.0, 80.0, 160.0, 90.0, 135.0, 15.0, 270.0, 30.0, 45.0,
                        15.0, 30.0, 45.0, 5.0, 30.0, 60.0, 90.0, 10.0, 60.0, 120.0, 180.0, 20.0, 120.0,
                        240.0, 360.0, 40.0, -20.0, -40.0, -80.0, -160.0, -59.0, -118.0, -178.5, -17.5]
        knap_constant = 563.0

        knapsack_prob_qubo = Knapsack(
            values, weights, weight_capacity, penalty).qubo

        self.assertTrue(terms_list_equality(
            knap_terms, knapsack_prob_qubo.terms))
        self.assertEqual(knap_weights, knapsack_prob_qubo.weights)
        self.assertEqual(knap_constant, knapsack_prob_qubo.constant)
        self.assertEqual(n_qubits, knapsack_prob_qubo.n)

    def test_knapsack_random_problem(self):
        """Test random instance method of Knapsack problem class"""

        rng = np.random.default_rng(1234)
        n_items = 5
        values = list(map(int, rng.integers(1, n_items, size=n_items)))
        weights = list(map(int, rng.integers(1, n_items, size=n_items)))
        weight_capacity = int(rng.integers(
            np.min(weights) * n_items, np.max(weights) * n_items))
        penalty = 2*np.max(values)

        knap_manual = Knapsack(
            values, weights, weight_capacity, int(penalty)).qubo

        knap_random_instance = Knapsack.random_instance(
            n_items=n_items, seed=1234).qubo

        self.assertTrue(terms_list_equality(
            knap_manual.terms, knap_random_instance.terms))
        self.assertEqual(knap_manual.weights, knap_random_instance.weights)
        self.assertEqual(knap_manual.constant, knap_random_instance.constant)
        self.assertEqual(knap_manual.n, knap_random_instance.n)

    def test_knapsack_random_problem_smallsize(self):
        """Test random instance method of Knapsack problem class"""

        rng = np.random.default_rng(1234)
        n_items = 3
        values = list(map(int, rng.integers(1, n_items, size=n_items)))
        weights = list(map(int, rng.integers(1, n_items, size=n_items)))
        weight_capacity = int(rng.integers(
            np.min(weights) * n_items, np.max(weights) * n_items))
        penalty = 2*np.max(values)

        knap_manual = Knapsack(
            values, weights, weight_capacity, int(penalty)).qubo

        knap_random_instance = Knapsack.random_instance(
            n_items=n_items, seed=1234).qubo

        self.assertTrue(terms_list_equality(
            knap_manual.terms, knap_random_instance.terms))
        self.assertEqual(knap_manual.weights, knap_random_instance.weights)
        self.assertEqual(knap_manual.constant, knap_random_instance.constant)
        self.assertEqual(knap_manual.n, knap_random_instance.n)

    def test_knapsack_type_checking(self):
        """
        Checks if the type-checking returns the right error.
        """

        # values type-check
        weights = [1, 2]
        weight_capacity = 5
        penalty = .1
        values_list = [(1, 2), {'test': 'oh', 'test1': 'oh'}, np.array([1, 2])]

        for each_values in values_list:
            with self.assertRaises(TypeError) as e:
                Knapsack(each_values, weights, weight_capacity, penalty)
            self.assertEqual("The input parameter, values, has to be a list",
                             str(e.exception))

        values_list = [['test', 'oh'], [np.array(1), np.array(2)], [.1, .5]]
        for each_values in values_list:
            with self.assertRaises(TypeError) as e:
                Knapsack(each_values, weights, weight_capacity, penalty)
            self.assertEqual("The elements in values list must be of type int.",
                             str(e.exception))

        # weights type-check
        values = [1, 2]
        weight_capacity = 5
        penalty = .1
        weights_list = [
            (1, 2), {'test': 'oh', 'test1': 'oh'}, np.array([1, 2])]

        for each_weights in weights_list:
            with self.assertRaises(TypeError) as e:
                Knapsack(values, each_weights, weight_capacity, penalty)
            self.assertEqual("The input parameter, weights, has to be a list",
                             str(e.exception))

        weights_list = [['test', 'oh'], [np.array(1), np.array(2)], [.1, .5]]
        for each_weights in weights_list:
            with self.assertRaises(TypeError) as e:
                Knapsack(values, each_weights, weight_capacity, penalty)
            self.assertEqual("The elements in weights list must be of type int.",
                             str(e.exception))

        # weight capacity type-check
        values = [1, 2]
        weights = [1, 2]
        weight_capacity_list = [.5, np.array(1), np.array(.5), 'oh']
        penalty = .1

        for each_weight_capacity in weight_capacity_list:
            with self.assertRaises(TypeError) as e:
                Knapsack(values, weights, each_weight_capacity, penalty)
            self.assertEqual("The input parameter, weight_capacity, has to be of type int",
                             str(e.exception))

        with self.assertRaises(TypeError) as e:
            Knapsack(values, weights, -1, penalty)
        self.assertEqual("The input parameter, weight_capacity, must be a positive integer greater than 0",
                         str(e.exception))

        # penalty capacity type-check
        values = [1, 2]
        weights = [1, 2]
        penalty_list = [np.array(1), np.array(.5), 'oh']
        weight_capacity = 5

        for each_penalty in penalty_list:
            with self.assertRaises(TypeError) as e:
                Knapsack(values, weights, weight_capacity, each_penalty)
            self.assertEqual("The input parameter, penalty, has to be of type float or int",
                             str(e.exception))

    # TESTING SLACKFREEKNAPSACK CLASS

    def test_slackfreeknapsack_terms_weights_constant(self):
        """Test that SlackFree Knapsack creates the correct QUBO problem"""

        values = [2, 4, 3, 5]
        weights = [3, 6, 9, 1]
        weight_capacity = 15
        n_qubits = len(values)
        penalty = 2*max(values)
        slknap_terms = [[0, 1], [0, 2], [0, 3], [
            1, 2], [1, 3], [2, 3], [0], [1], [2], [3]]
        slknap_weights = [90.0, 135.0, 15.0, 270.0,
                          30.0, 45.0, 166.0, 332.0, 496.5, 57.5]
        slknap_constant = 613.0

        slknapsack_prob_qubo = SlackFreeKnapsack(
            values, weights, weight_capacity, penalty).qubo

        self.assertTrue(terms_list_equality(
            slknap_terms, slknapsack_prob_qubo.terms))
        self.assertEqual(slknap_weights, slknapsack_prob_qubo.weights)
        self.assertEqual(slknap_constant, slknapsack_prob_qubo.constant)
        self.assertEqual(n_qubits, slknapsack_prob_qubo.n)

    def test_slackfreeknapsack_random_problem(self):
        """Test random instance method of SlackFree Knapsack problem class"""

        rng = np.random.default_rng(1234)
        n_items = 5
        values = list(map(int, rng.integers(1, n_items, size=n_items)))
        weights = list(map(int, rng.integers(1, n_items, size=n_items)))
        weight_capacity = int(rng.integers(
            np.min(weights) * n_items, np.max(weights) * n_items))
        penalty = 2*np.max(values)

        slknap_manual = SlackFreeKnapsack(
            values, weights, weight_capacity, int(penalty)).qubo

        slknap_random_instance = SlackFreeKnapsack.random_instance(
            n_items=n_items, seed=1234).qubo

        self.assertTrue(terms_list_equality(
            slknap_manual.terms, slknap_random_instance.terms))
        self.assertEqual(slknap_manual.weights, slknap_random_instance.weights)
        self.assertEqual(slknap_manual.constant,
                         slknap_random_instance.constant)
        self.assertEqual(slknap_manual.n, slknap_random_instance.n)

    # TESTING MINIMUMVERTEXCOVER CLASS

    def test_mvc_terms_weights_constant(self):
        """Test terms,weights,constant of QUBO generated by MVC class"""

        mvc_terms = [[0, 3], [0, 4], [1, 2], [1, 3], [2, 4], [3, 4],
                     [0], [1], [2], [3], [4]]
        mvc_weights = [1.25, 1.25, 1.25, 1.25,
                       1.25, 1.25, 2.0, 2.0, 2.0, 3.25, 3.25]
        mvc_constant = 10.0

        gr = nx.generators.fast_gnp_random_graph(5, 0.8, seed=1234)
        mvc_prob = MinimumVertexCover(
            gr, field=1.0, penalty=5).qubo

        self.assertTrue(terms_list_equality(mvc_terms, mvc_prob.terms))
        self.assertEqual(mvc_weights, mvc_prob.weights)
        self.assertEqual(mvc_constant, mvc_prob.constant)

    def test_mvc_random_problem(self):
        """Test the random_instance method of MVC class"""
        mvc_terms = [[0, 3], [0, 4], [1, 2], [1, 3],
                     [2, 4], [3, 4], [0], [1], [2], [3], [4]]
        mvc_weights = [2.5, 2.5, 2.5, 2.5, 2.5, 2.5, 4.5, 4.5, 4.5, 7.0, 7.0]
        mvc_constant = 17.5

        mvc_prob_random = MinimumVertexCover.random_instance(
            n_nodes=5, edge_probability=0.8, seed=1234).qubo

        self.assertTrue(terms_list_equality(mvc_terms, mvc_prob_random.terms))
        self.assertEqual(mvc_weights, mvc_prob_random.weights)
        self.assertEqual(mvc_constant, mvc_prob_random.constant)

    def test_mvc_type_checking(self):
        """
        Checks if the type-checking returns the right error.
        """

        # graph type-check
        graph_list = [(1, 2), {'node1': 1, 'node2': 2}, np.array([1, 2])]
        field = .1
        penalty = .1

        for each_graph in graph_list:
            with self.assertRaises(TypeError) as e:
                MinimumVertexCover(each_graph, field, penalty)
            self.assertEqual("Input problem graph must be a networkx Graph.",
                             str(e.exception))

        # field capacity type-check
        graph = nx.circulant_graph(6, [1])
        field_list = [np.array(1), np.array(.5), 'oh']
        penalty = .1

        for each_field in field_list:
            with self.assertRaises(TypeError) as e:
                MinimumVertexCover(graph, each_field, penalty)
            self.assertEqual("The input parameter, field, has to be of type float or int",
                             str(e.exception))

        # penalty capacity type-check
        graph = nx.circulant_graph(6, [1])
        field = .1
        penalty_list = [np.array(1), np.array(.5), 'oh']

        for each_penalty in penalty_list:
            with self.assertRaises(TypeError) as e:
                MinimumVertexCover(graph, field, each_penalty)
            self.assertEqual("The input parameter, penalty, has to be of type float or int",
                             str(e.exception))

    # TESTING TSP PROBLEM CLASS

    def test_tsp_terms_weights_constant(self):
        """Testing TSP problem creation"""
        city_coordinates = [(4, 1), (4, 4), (3, 3), (1, 3.5)]
        expected_terms = [[0, 3],
                          [1, 4],
                          [2, 5],
                          [0, 6],
                          [1, 7],
                          [8, 2],
                          [3, 6],
                          [4, 7],
                          [8, 5],
                          [0, 1],
                          [0, 2],
                          [1, 2],
                          [3, 4],
                          [3, 5],
                          [4, 5],
                          [6, 7],
                          [8, 6],
                          [8, 7],
                          [0, 4],
                          [1, 3],
                          [3, 7],
                          [4, 6],
                          [0, 5],
                          [2, 3],
                          [8, 3],
                          [5, 6],
                          [1, 5],
                          [2, 4],
                          [8, 4],
                          [5, 7],
                          [0],
                          [1],
                          [2],
                          [3],
                          [4],
                          [5],
                          [6],
                          [7],
                          [8]]
        expected_weights = [3.905124837953327,
                            3.905124837953327,
                            3.905124837953327,
                            3.905124837953327,
                            3.905124837953327,
                            3.905124837953327,
                            3.905124837953327,
                            3.905124837953327,
                            3.905124837953327,
                            3.905124837953327,
                            3.905124837953327,
                            3.905124837953327,
                            3.905124837953327,
                            3.905124837953327,
                            3.905124837953327,
                            3.905124837953327,
                            3.905124837953327,
                            3.905124837953327,
                            0.3535533905932738,
                            0.3535533905932738,
                            0.7603453162872774,
                            0.7603453162872774,
                            0.3535533905932738,
                            0.3535533905932738,
                            0.5153882032022076,
                            0.5153882032022076,
                            0.7603453162872774,
                            0.7603453162872774,
                            0.5153882032022076,
                            0.5153882032022076,
                            -10.424148382787205,
                            -9.797225258452029,
                            -11.038545614372802,
                            -10.038047089667756,
                            -9.548132863497614,
                            -10.361716714885624,
                            -10.424148382787205,
                            -9.797225258452029,
                            -11.038545614372802]
        expected_constant = 62.51983851122417
        tsp_qubo = TSP(city_coordinates).qubo
        print(tsp_qubo.weights)
        self.assertTrue(terms_list_equality(expected_terms, tsp_qubo.terms))
        self.assertEqual(expected_weights, tsp_qubo.weights)
        self.assertEqual(expected_constant, tsp_qubo.constant)

    def test_tsp_random_instance(self):
        """Testing the random_instance method of the TSP problem class"""
        rng = np.random.default_rng(1234)
        n_cities = 4

        box_size = np.sqrt(n_cities)
        city_coordinates = list(
            map(tuple, box_size * rng.random(size=(n_cities, 2))))

        tsp_prob = TSP(city_coordinates).qubo

        tsp_prob_random = TSP.random_instance(
            n_cities=n_cities, seed=1234).qubo

        self.assertTrue(terms_list_equality(
            tsp_prob_random.terms, tsp_prob.terms))
        self.assertEqual(tsp_prob_random.weights, tsp_prob.weights)
        self.assertEqual(tsp_prob_random.constant, tsp_prob.constant)

    def test_tsp_type_checking(self):
        """
        Checks if the type-checking returns the right error.
        """
        # If nothing is given, must return a ValueError
        with self.assertRaises(ValueError) as e:
            TSP()
        self.assertEqual("Input missing: city coordinates, distance matrix or (weighted graph) required",
                         str(e.exception))

        # coordinates type-check
        coordinates_list = [
            (1, 2), {'test': 'oh', 'test1': 'oh'}, np.array([1, 2])]

        for each_coordinates in coordinates_list:
            with self.assertRaises(TypeError) as e:
                TSP(each_coordinates)
            self.assertEqual("The coordinates should be a list",
                             str(e.exception))

        coordinates_list = [[[1, 2], [2, 1]], [
            np.array([1, 2]), np.array([2, 1])]]
        for each_coordinates in coordinates_list:
            with self.assertRaises(TypeError) as e:
                TSP(each_coordinates)
            self.assertEqual("The coordinates should be contained in a tuple",
                             str(e.exception))

        coordinates_list = [[('oh', 'num'), ('num', 'oh')],
                            [(np.array(1), np.array(2)), (np.array(2), np.array(1))]]
        for each_coordinates in coordinates_list:
            with self.assertRaises(TypeError) as e:
                TSP(each_coordinates)
            self.assertEqual("The coordinates must be of type float or int",
                             str(e.exception))

        # coordinates type-check
        distance_matrices = [(1, 2), np.array([[1, 2], [3, 4]]), {
            'test': 'oh', 'test1': 'oh'}]

        for distance_matrix in distance_matrices:
            with self.assertRaises(TypeError) as e:
                TSP(distance_matrix=distance_matrix)
            self.assertEqual("The distance matrix should be a list",
                             str(e.exception))

        # Distance matrix type-check
        distance_matrices = [[(1, 2), (2, 1)], [
            np.array([1, 2]), np.array([2, 1])]]
        for distance_matrix in distance_matrices:
            with self.assertRaises(TypeError) as e:
                TSP(distance_matrix=distance_matrix)
            self.assertEqual("Each row in the distance matrix should be a list",
                             str(e.exception))

        distance_matrices = [[['oh', 'num'], ['num', 'oh']],
                             [[np.array(1), np.array(2)], [np.array(2), np.array(1)]]]
        for distance_matrix in distance_matrices:
            with self.assertRaises(TypeError) as e:
                TSP(distance_matrix=distance_matrix)
            self.assertEqual("The distance matrix entries must be of type float or int",
                             str(e.exception))

        distance_matrix = [[1, 2.3], [-2, 3]]
        with self.assertRaises(ValueError) as e:
            TSP(distance_matrix=distance_matrix)
        self.assertEqual("Distances should be positive",
                         str(e.exception))

        # Graph type-check
        G = nx.complete_graph(5)
        for (u, v) in G.edges():
            G[u][v]['weight'] = 'a'

        with self.assertRaises(TypeError) as e:
            TSP(G=G)
        self.assertEqual("The edge weights must be of type float or int",
                         str(e.exception))

        for (u, v) in G.edges():
            G[u][v]['weight'] = -2.

        with self.assertRaises(ValueError) as e:
            TSP(G=G)
        self.assertEqual("Edge weights should be positive",
                         str(e.exception))

    # TESTING SHORTESTPATH PROBLEM CLASS

    def test_shortestpath_terms_weights_constant(self):
        """Test terms,weights,constant of QUBO generated by Shortest Path class"""

        sp_terms = [[0], [1], [2], [3], [1], [1, 2], [2, 1], [2], [2], [2, 3],
                    [3, 2], [3], [0], [0, 1], [1], [1, 3], [0, 3], [3, 1], [3]]
        sp_weights = [1, 1, 1, 1, -1, 1, 1, -
                      1, -1, 1, 1, -1, 4, -4, 1, 1, -4, 1, 1]
        conv_sp_terms = [[1, 2], [2, 1], [2, 3], [3, 2], [
            0, 1], [1, 3], [0, 3], [3, 1], [0], [1], [2], [3], []]
        conv_sp_weights = [0.25, 0.25, 0.25, 0.25, -1.0,
                           0.25, -1.0, 0.25, -0.5, -0.5, -0.5, -0.5, 2.5]
        sp_qubo_terms = [[1, 2], [2, 3], [0, 1],
                         [1, 3], [0, 3], [0], [1], [2], [3]]
        sp_qubo_weights = [0.5, 0.5, -1.0, 0.5, -1.0, -0.5, -0.5, -0.5, -0.5]
        sp_qubo_constant = 2.5

        gr = nx.generators.fast_gnp_random_graph(3, 1, seed=1234)
        for (u, v) in gr.edges():
            gr.edges[u, v]['weight'] = 1
        for w in gr.nodes():
            gr.nodes[w]['weight'] = 1
        source, dest = 0, 2
        sp = ShortestPath(gr, source, dest)
        n_variables = sp.G.number_of_nodes() + sp.G.number_of_edges() - 2
        bin_terms, bin_weights = sp.terms_and_weights()
        terms, weights = QUBO.convert_qubo_to_ising(
            n_variables, bin_terms, bin_weights)
        qubo = sp.qubo
        print(terms)
        self.assertTrue(terms_list_equality(bin_terms, sp_terms))
        self.assertEqual(list(bin_weights), sp_weights)
        self.assertTrue(terms_list_equality(terms, conv_sp_terms))
        self.assertEqual(list(weights), conv_sp_weights)
        self.assertTrue(terms_list_equality(sp_qubo_terms, qubo.terms))
        self.assertEqual(sp_qubo_weights, qubo.weights)
        self.assertEqual(sp_qubo_constant, qubo.constant)

    def test_shortestpath_random_instance(self):
        """Test random instance method of Shortest Path problem class"""
        sp_rand_terms = [[1, 2], [2, 3], [0, 1],
                         [1, 3], [0, 3], [0], [1], [2], [3]]
        sp_rand_weights = [0.5, 0.5, -1.0, 0.5, -1.0, -0.5, -0.5, -0.5, -0.5]
        sp_rand_constant = 2.5

        gr = nx.generators.fast_gnp_random_graph(3, 1, seed=1234)
        for (u, v) in gr.edges():
            gr.edges[u, v]['weight'] = 1.0
        for w in gr.nodes():
            gr.nodes[w]['weight'] = 1.0
        sp_prob = ShortestPath.random_instance(
            n_nodes=3, edge_probability=1, seed=1234, source=0, dest=2).qubo
        print(sp_prob.terms)
        self.assertTrue(terms_list_equality(sp_rand_terms, sp_prob.terms))
        self.assertEqual(sp_rand_weights, sp_prob.weights)
        self.assertEqual(sp_rand_constant, sp_prob.constant)

        self.assertEqual(sp_prob.terms, ShortestPath(
            gr, 0, 2).qubo.terms)
        self.assertEqual(sp_prob.weights, ShortestPath(
            gr, 0, 2).qubo.weights)
        self.assertEqual(sp_prob.constant, ShortestPath(
            gr, 0, 2).qubo.constant)

    def test_assertion_error(self):

        def test_assertion_fn():
            n_row = 1
            n_col = 1

            G = nx.triangular_lattice_graph(n_row, n_col)
            G = nx.convert_node_labels_to_integers(G)
            G.remove_edges_from(nx.selfloop_edges(G))

            node_weights = np.round(np.random.rand(len(G.nodes())), 3)
            edge_weights = np.round(np.random.rand(len(G.edges())), 3)

            node_dict = dict(zip(list(G.nodes()), node_weights))
            edge_dict = dict(zip(list(G.edges()), edge_weights))

            nx.set_edge_attributes(G, values=edge_dict, name='weight')
            nx.set_node_attributes(G, values=node_dict, name='weight')

            shortest_path_problem = ShortestPath(G, 0, -1)
            shortest_path_qubo = shortest_path_problem.qubo

        self.assertRaises(Exception, test_assertion_fn)

    def __generate_random_problems(self):
        problems_random_instances = {
            "tsp":TSP.random_instance(n_cities=randint(2, 15)),
            "number_partition":NumberPartition.random_instance(n_numbers=randint(2, 15)),
            "maximum_cut":MaximumCut.random_instance(n_nodes=randint(2, 15), edge_probability=random()),
            "knapsack":Knapsack.random_instance(n_items=randint(2, 15)),
            "slack_free_knapsack":SlackFreeKnapsack.random_instance(n_items=randint(2, 15)),
            "minimum_vertex_cover":MinimumVertexCover.random_instance(n_nodes=randint(2, 15), edge_probability=random()),
            "shortest_path":ShortestPath.random_instance(n_nodes=randint(3, 15), edge_probability=random()),
        }
        qubo_random_instances = {k:v.qubo for k,v in problems_random_instances.items()}
        qubo_random_instances["generic_qubo"] = QUBO.random_instance(randint(2, 15))

        return problems_random_instances, qubo_random_instances

    def test_problem_instance(self):
        """
        Test problem instance method of the QUBO class.
        From the random instance of all the different problems, we generate the QUBO problem out of it and then we check if the problem instance attribute is correct, by comparing the keys of the problem instance with the expected keys.
        """

        _, qubos = self.__generate_random_problems()

        expected_keys = {
            "tsp":['problem_type', 'n_cities', 'G', 'A', 'B'],
            "number_partition":['problem_type', 'numbers', 'n_numbers'],
            "maximum_cut":['problem_type', 'G'],
            "knapsack":['problem_type', 'values', 'weights', 'weight_capacity', 'penalty', 'n_items'],
            "slack_free_knapsack":['problem_type', 'values', 'weights', 'weight_capacity', 'penalty', 'n_items'],
            "minimum_vertex_cover":['problem_type', 'G', 'field', 'penalty'],
            "shortest_path":['problem_type', 'G', 'source', 'dest'],
            "generic_qubo":['problem_type']
        }

        for k,v in qubos.items():
            assert list(v.problem_instance.keys()) == expected_keys[k], "Problem instance keys are not correct for problem type {}".format(k)
            assert k == v.problem_instance['problem_type'], "Problem type is not correct for problem type {}".format(k)


    def test_problem_from_instance_dict(self):
        """
        Test problem from instance method of the problem class.
        """
        problem_mapper = {
            "generic_qubo": QUBO,
            "tsp": TSP,
            "number_partition": NumberPartition,
            "maximum_cut": MaximumCut,
            "knapsack": Knapsack,
            "slack_free_knapsack": SlackFreeKnapsack,
            "minimum_vertex_cover": MinimumVertexCover,
            "shortest_path": ShortestPath,
        }

        problems, qubos = self.__generate_random_problems()

        for type in qubos:
            if type == "generic_qubo":
                continue

            problem_instance = qubos[type].problem_instance.copy()

            problem = create_problem_from_dict(problem_instance)

            assert problem.problem_instance == problems[type].problem_instance, "Problem from instance method is not correct for problem type {}".format(type)
            assert convert2serialize(problem) == convert2serialize(problems[type]), "Problem from instance method is not correct for problem type {}".format(type)

        
    def test_qubo_from_dict(self):
        """
        Test qubo from dict method of the QUBO class.
        """

        _, qubos = self.__generate_random_problems()
        for _, qubo in qubos.items():

            qubo_dict = qubo.asdict()

            new_qubo = QUBO.from_dict(qubo_dict)

            for term, new_term in zip(qubo.terms, new_qubo.terms):
                assert set(term) == set(new_term), "QUBO from dict method is not correct for problem type {}, terms compared: {}, {}".format(qubo.problem_instance['problem_type'], term, new_term)

                assert set(qubo.weights) == set(new_qubo.weights), "QUBO from dict method is not correct for problem type {}".format(qubo.problem_instance['problem_type'])

                for key in qubo.__dict__:
                    if key != "terms" and key != "weights":
                        assert qubo.__dict__[key] == new_qubo.__dict__[key], "QUBO from dict method is not correct for problem type {}".format(qubo.problem_instance['problem_type'])
        


if __name__ == '__main__':
    unittest.main()



import unittest
import numpy as np

from qiskit import QuantumCircuit
from pyquil import Program, quilbase
from pyquil.gates import RX as p_RX
from pyquil.gates import RY as p_RY
from pyquil.gates import RZ as p_RZ
from pyquil.gates import CZ as p_CZ
from pyquil.gates import CNOT as p_CX
from pyquil.gates import XY as p_XY
from pyquil.gates import CPHASE as p_CPHASE
from braket.circuits import gates as braketgates
from braket.circuits import Circuit
from braket.circuits.free_parameter import FreeParameter

from openqaoa.qaoa_components.ansatz_constructor.gates import (RY, RX, RZ, CZ, CX, 
                                                               RXX, RYY, RZZ, RZX, 
                                                               CPHASE, RiSWAP)
from openqaoa.qaoa_components.ansatz_constructor.rotationangle import RotationAngle

class TestingGate(unittest.TestCase):
    
    def test_braket_gates_1q(self):
        
        # One Qubit Gate Tests
        rotation_angle_obj = RotationAngle(lambda x: x, [], FreeParameter('test_angle'))
        
        empty_circuit = Circuit()
        llgate = RY()
        output_circuit = llgate.apply_braket_gate(0,rotation_angle_obj,empty_circuit)
        output_circuit = output_circuit.make_bound_circuit({'test_angle': np.pi})
        
        test_circuit = Circuit()
        test_circuit += braketgates.Ry.ry(0,np.pi)
        
        self.assertEqual(test_circuit, output_circuit)
        
        empty_circuit = Circuit()
        llgate = RX()
        output_circuit = llgate.apply_braket_gate(0,rotation_angle_obj,empty_circuit)
        output_circuit = output_circuit.make_bound_circuit({'test_angle': np.pi})
        
        test_circuit = Circuit()
        test_circuit += braketgates.Rx.rx(0,np.pi)
        
        self.assertEqual(test_circuit, output_circuit)
        
        empty_circuit = Circuit()
        llgate = RZ()
        output_circuit = llgate.apply_braket_gate(0,rotation_angle_obj,empty_circuit)
        output_circuit = output_circuit.make_bound_circuit({'test_angle': np.pi})
        
        test_circuit = Circuit()
        test_circuit += braketgates.Rz.rz(0,np.pi)
        
        self.assertEqual(test_circuit, output_circuit)
        
    def test_braket_gates_2q(self):
        
        # Two Qubit Gate Tests
        empty_circuit = Circuit()
        llgate = CZ()
        output_circuit = llgate.apply_braket_gate([0, 1],empty_circuit)

        test_circuit = Circuit()
        test_circuit += braketgates.CZ.cz(0, 1)
        
        self.assertEqual(test_circuit, output_circuit)
        
        empty_circuit = Circuit()
        llgate = CX(mode='CX')
        output_circuit = llgate.apply_braket_gate([0, 1],empty_circuit)

        test_circuit = Circuit()
        test_circuit += braketgates.CNot.cnot(0, 1)
        
        self.assertEqual(test_circuit, output_circuit)
        
    def test_braket_gates_2q_w_gates(self):
        
        # Two Qubit Gate with Angles Tests
        rotation_angle_obj = RotationAngle(lambda x: x, [], FreeParameter('test_angle'))
        
        empty_circuit = Circuit()
        llgate = RXX()
        output_circuit = llgate.apply_braket_gate([0, 1], rotation_angle_obj, empty_circuit)
        output_circuit = output_circuit.make_bound_circuit({'test_angle': np.pi})
        
        test_circuit = Circuit()
        test_circuit += braketgates.XX.xx(0, 1, np.pi)
        
        self.assertEqual(test_circuit, output_circuit)
        
        empty_circuit = Circuit()
        llgate = RYY()
        output_circuit = llgate.apply_braket_gate([0, 1], rotation_angle_obj, empty_circuit)
        output_circuit = output_circuit.make_bound_circuit({'test_angle': np.pi})
        
        test_circuit = Circuit()
        test_circuit += braketgates.YY.yy(0, 1, np.pi)
        
        self.assertEqual(test_circuit, output_circuit)
        
        empty_circuit = Circuit()
        llgate = RZZ()
        output_circuit = llgate.apply_braket_gate([0, 1], rotation_angle_obj, empty_circuit)
        output_circuit = output_circuit.make_bound_circuit({'test_angle': np.pi})
        
        test_circuit = Circuit()
        test_circuit += braketgates.ZZ.zz(0, 1, np.pi)
        
        self.assertEqual(test_circuit, output_circuit)
        
        empty_circuit = Circuit()
        llgate = CPHASE()
        output_circuit = llgate.apply_braket_gate([0, 1], rotation_angle_obj, empty_circuit)
        output_circuit = output_circuit.make_bound_circuit({'test_angle': np.pi})
        
        test_circuit = Circuit()
        test_circuit += braketgates.CPhaseShift.cphaseshift(0, 1, np.pi)
        
        self.assertEqual(test_circuit, output_circuit)
        
        empty_circuit = Circuit()
        llgate = RiSWAP()
        output_circuit = llgate.apply_braket_gate([0, 1], rotation_angle_obj, empty_circuit)
        output_circuit = output_circuit.make_bound_circuit({'test_angle': np.pi})
        
        test_circuit = Circuit()
        test_circuit += braketgates.XY.xy(0, 1, np.pi)
        
        self.assertEqual(test_circuit, output_circuit)
    
    def test_ibm_gates_1q(self):
        
        # One Qubit Gate Tests
        rotation_angle_obj = RotationAngle(lambda x: x, [], np.pi)
        
        empty_circuit = QuantumCircuit(1)
        llgate = RY()
        output_circuit = llgate.apply_ibm_gate(0,rotation_angle_obj,empty_circuit)

        test_circuit = QuantumCircuit(1)
        test_circuit.ry(np.pi, 0)
        
        self.assertEqual(test_circuit.to_instruction().definition, output_circuit.to_instruction().definition)
        
        empty_circuit = QuantumCircuit(1)
        llgate = RX()
        output_circuit = llgate.apply_ibm_gate(0,rotation_angle_obj,empty_circuit)

        test_circuit = QuantumCircuit(1)
        test_circuit.rx(np.pi, 0)
        
        self.assertEqual(test_circuit.to_instruction().definition, output_circuit.to_instruction().definition)
        
        empty_circuit = QuantumCircuit(1)
        llgate = RZ()
        output_circuit = llgate.apply_ibm_gate(0, rotation_angle_obj, empty_circuit)

        test_circuit = QuantumCircuit(1)
        test_circuit.rz(np.pi, 0)
        
        self.assertEqual(test_circuit.to_instruction().definition, output_circuit.to_instruction().definition)
        
    def test_ibm_gates_2q(self):
            
        # Two Qubit Gate Tests
        empty_circuit = QuantumCircuit(2)
        llgate = CZ()
        output_circuit = llgate.apply_ibm_gate([0, 1],empty_circuit)

        test_circuit = QuantumCircuit(2)
        test_circuit.cz(0, 1)
        
        self.assertEqual(test_circuit.to_instruction().definition, output_circuit.to_instruction().definition)
        
        empty_circuit = QuantumCircuit(2)
        llgate = CX(mode='CX')
        output_circuit = llgate.apply_ibm_gate([0, 1],empty_circuit)

        test_circuit = QuantumCircuit(2)
        test_circuit.cx(0, 1)
        
        self.assertEqual(test_circuit.to_instruction().definition, output_circuit.to_instruction().definition)
        
        empty_circuit = QuantumCircuit(2)
        llgate = CX(mode='CZ')
        output_circuit = llgate.apply_ibm_gate([0, 1], empty_circuit)

        test_circuit = QuantumCircuit(2)
        test_circuit.ry(np.pi/2, 1)
        test_circuit.rx(np.pi, 1)
        test_circuit.cz(0, 1)
        test_circuit.ry(np.pi/2, 1)
        test_circuit.rx(np.pi, 1)
        
        self.assertEqual(test_circuit.to_instruction().definition, output_circuit.to_instruction().definition)
        
    def test_ibm_gates_2q_w_gates(self):
        
        # Two Qubit Gate with Angles Tests
        rotation_angle_obj = RotationAngle(lambda x: x, [], np.pi)
        
        empty_circuit = QuantumCircuit(2)
        llgate = RXX()
        output_circuit = llgate.apply_ibm_gate([0, 1], rotation_angle_obj, empty_circuit)

        test_circuit = QuantumCircuit(2)
        test_circuit.rxx(np.pi, 0, 1)
        
        self.assertEqual(test_circuit.to_instruction().definition, output_circuit.to_instruction().definition)
        
        empty_circuit = QuantumCircuit(2)
        llgate = RYY()
        output_circuit = llgate.apply_ibm_gate([0, 1], rotation_angle_obj, empty_circuit)

        test_circuit = QuantumCircuit(2)
        test_circuit.ryy(np.pi, 0, 1)
        
        self.assertEqual(test_circuit.to_instruction().definition, output_circuit.to_instruction().definition)
        
        empty_circuit = QuantumCircuit(2)
        llgate = RZZ()
        output_circuit = llgate.apply_ibm_gate([0, 1], rotation_angle_obj, empty_circuit)

        test_circuit = QuantumCircuit(2)
        test_circuit.rzz(np.pi, 0, 1)
        
        self.assertEqual(test_circuit.to_instruction().definition, output_circuit.to_instruction().definition)
        
        empty_circuit = QuantumCircuit(2)
        llgate = RZX()
        output_circuit = llgate.apply_ibm_gate([0, 1], rotation_angle_obj, empty_circuit)

        test_circuit = QuantumCircuit(2)
        test_circuit.rzx(np.pi, 0, 1)
        
        self.assertEqual(test_circuit.to_instruction().definition, output_circuit.to_instruction().definition)
        
        empty_circuit = QuantumCircuit(2)
        llgate = CPHASE()
        output_circuit = llgate.apply_ibm_gate([0, 1], rotation_angle_obj, empty_circuit)

        test_circuit = QuantumCircuit(2)
        test_circuit.crz(np.pi, 0, 1)
        
        self.assertEqual(test_circuit.to_instruction().definition, output_circuit.to_instruction().definition)
        
    def test_pyquil_gates_1q(self):
        
        # One Qubit Gate Tests
        rotation_angle_obj = RotationAngle(lambda x: x, [], np.pi)
        
        empty_program = Program()
        llgate = RY()
        output_program = llgate.apply_pyquil_gate(0,rotation_angle_obj,empty_program)

        test_program = Program().inst(p_RY(np.pi, 0))
        
        output_gate_names = [instr.name for instr in output_program if type(instr) == quilbase.Gate]
        test_gate_names = [instr.name for instr in test_program if type(instr) == quilbase.Gate]
        self.assertEqual(output_gate_names, test_gate_names)
        
        empty_program = Program()
        llgate = RX()
        output_program = llgate.apply_pyquil_gate(0,rotation_angle_obj,empty_program)

        test_program = Program().inst(p_RX(np.pi, 0))
        
        output_gate_names = [instr.name for instr in output_program if type(instr) == quilbase.Gate]
        test_gate_names = [instr.name for instr in test_program if type(instr) == quilbase.Gate]
        self.assertEqual(output_gate_names, test_gate_names)
        
        empty_program = Program()
        llgate = RZ()
        output_program = llgate.apply_pyquil_gate(0,rotation_angle_obj,empty_program)
        
        test_program = Program().inst(p_RZ(np.pi, 0))
        
        output_gate_names = [instr.name for instr in output_program if type(instr) == quilbase.Gate]
        test_gate_names = [instr.name for instr in test_program if type(instr) == quilbase.Gate]
        self.assertEqual(output_gate_names, test_gate_names)
        
    def test_pyquil_gates_2q(self):
        
        # Two Qubit Gate Tests
        empty_program = Program()
        llgate = CZ()
        output_program = llgate.apply_pyquil_gate([0, 1],empty_program)

        test_program = Program().inst(p_CZ(0, 1))
        
        output_gate_names = [instr.name for instr in output_program if type(instr) == quilbase.Gate]
        test_gate_names = [instr.name for instr in test_program if type(instr) == quilbase.Gate]
        self.assertEqual(output_gate_names, test_gate_names)
        
        empty_program = Program()
        llgate = CX()
        output_program = llgate.apply_pyquil_gate([0, 1], empty_program)

        test_program = Program().inst(p_CX(0, 1))
        
        output_gate_names = [instr.name for instr in output_program if type(instr) == quilbase.Gate]
        test_gate_names = [instr.name for instr in test_program if type(instr) == quilbase.Gate]
        self.assertEqual(output_gate_names, test_gate_names)
        
    def test_pyquil_gates_2q_w_gates(self):
        
        # Two Qubit Gate with Angles Tests
        rotation_angle_obj = RotationAngle(lambda x: x, [], np.pi)
        
        empty_program = Program()
        llgate = CPHASE()
        output_program = llgate.apply_pyquil_gate([0, 1], rotation_angle_obj, empty_program)

        test_program = Program().inst(p_CPHASE(np.pi, 0, 1))
        
        output_gate_names = [instr.name for instr in output_program if type(instr) == quilbase.Gate]
        test_gate_names = [instr.name for instr in test_program if type(instr) == quilbase.Gate]
        self.assertEqual(output_gate_names, test_gate_names)
        
        empty_program = Program()
        llgate = RiSWAP()
        output_program = llgate.apply_pyquil_gate([0, 1], rotation_angle_obj, empty_program)

        test_program = Program().inst(p_XY(np.pi, 0, 1))
        
        output_gate_names = [instr.name for instr in output_program if type(instr) == quilbase.Gate]
        test_gate_names = [instr.name for instr in test_program if type(instr) == quilbase.Gate]
        self.assertEqual(output_gate_names, test_gate_names)
    

if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from abc import ABC
from typing import List, Union, Tuple, Any, Callable, Iterable
import numpy as np

from .operators import Hamiltonian
from .hamiltonianmapper import HamiltonianMapper

from .gatemap import RotationGateMap

def _is_iterable_empty(in_iterable):
    if isinstance(in_iterable, Iterable):  # Is Iterable
        return all(map(_is_iterable_empty, in_iterable))
    return False  # Not an Iterable


class shapedArray:
    """Decorator-Descriptor for arrays that have a fixed shape.

    This is used to facilitate automatic input checking for all the different
    internal parameters. Each instance of this class can be removed without
    replacement and the code should still work, provided the user provides
    only correct angles to below parameter classes

    Parameters
    ----------
    shape: `Callable[[Any], Tuple]`
        Returns the shape for self.values

    Example
    -------
    With this descriptor, the two following are equivalent:

    .. code-block:: python

        class foo():
            def __init__(self):
                self.shape = (n, m)
                self._my_attribute = None

            @property
            def my_attribute(self):
                return _my_attribute

            @my_attribute.setter
            def my_attribute(self):
                try:
                    self._my_attribute = np.reshape(values, self.shape)
                except ValueError:
                    raise ValueError("my_attribute must have shape "
                                    f"{self.shape}")


    can be simplified to

    .. code-block:: python

        class foo():
            def __init__(self):
                self.shape = (n, m)

            @shapedArray
            def my_attribute(self):
                return self.shape
    """

    def __init__(self, shape: Callable[[Any], Tuple]):
        """The constructor. See class documentation"""
        self.name = shape.__name__
        self.shape = shape

    def __set__(self, obj, values):
        """The setter with input checking."""
        try:
            # also round to 12 decimal places to avoid floating point errors
            setattr(
                obj, f"__{self.name}", np.round(np.reshape(values, self.shape(obj)), 12)
            )
        except ValueError:
            raise ValueError(f"{self.name} must have shape {self.shape(obj)}")

    def __get__(self, obj, objtype):
        """The getter."""
        return getattr(obj, f"__{self.name}")


class AnsatzDescriptor(ABC):
    """
    Parameters class to construct a specific quantum ansatz to attack
    a problem

    Parameters
    ----------
    algorithm: `str`

    Attributes
    ----------
    algorithm: `str`
    """

    def __init__(self, algorithm: str):
        self.algorithm = algorithm


class QAOADescriptor(AnsatzDescriptor):

    """
    Create the problem attributes consisting of the Hamiltonian, QAOA 'p'
    value and other specific parameters.

    Parameters
    ----------
    cost_hamiltonian: `Hamiltonian`
        The cost hamiltonian of the problem the user is trying to solve.

    mixer_block: `Union[List[RotationGateMap], Hamiltonian]`
        The mixer hamiltonian or a list of initialised RotationGateMap objects
        that defines the gates to be used within the "mixer part" of the circuit.

    p: `int`
        Number of QAOA layers; defaults to 1 if not specified

    mixer_coeffs: `List[float]`
        A list containing coefficients for each mixer GateMap. The order of the
        coefficients should follow the order of the GateMaps provided in the relevant gate block.
        This input isnt required if the input mixer block is of type Hamiltonian.

    Attributes
    ----------
    cost_hamiltonian: `Hamiltonian`

    qureg: `List[int]`

    cost_block_coeffs: `List[float]`

    cost_single_qubit_coeffs: `List[float]`

    cost_qubits_singles: `List[str]`

    cost_pair_qubit_coeffs: `List[float]`

    cost_qubits_pairs: `List[str]`

    mixer_block_coeffs: `List[float]`

    Properties
    ----------
    cost_block: `List[RotationGateMap]`

    mixer_block: `List[RotationGateMap]`

    abstract_circuit: `List[RotationGateMap]`
    """

    def __init__(
        self,
        cost_hamiltonian: Hamiltonian,
        mixer_block: Union[List[RotationGateMap], Hamiltonian],
        p: int,
        mixer_coeffs: List[float] = [],
    ):

        super().__init__(algorithm="QAOA")

        self.p = p

        self.qureg = cost_hamiltonian.qureg
        self.cost_block_coeffs = cost_hamiltonian.coeffs

        try:
            self.mixer_block_coeffs = mixer_block.coeffs
        except AttributeError:
            self.mixer_block_coeffs = mixer_coeffs

        # Needed in the BaseBackend to compute exact_solution, cost_funtion method
        # and bitstring_energy
        self.cost_hamiltonian = cost_hamiltonian
        self.cost_block = cost_hamiltonian
        self.mixer_block = mixer_block

        (
            self.cost_single_qubit_coeffs,
            self.cost_pair_qubit_coeffs,
            self.cost_qubits_singles,
            self.cost_qubits_pairs,
        ) = self._assign_coefficients(self.cost_block[0], self.cost_block_coeffs)

        (
            self.mixer_single_qubit_coeffs,
            self.mixer_pair_qubit_coeffs,
            self.mixer_qubits_singles,
            self.mixer_qubits_pairs,
        ) = self._assign_coefficients(self.mixer_block[0], self.mixer_block_coeffs)

    def _assign_coefficients(
        self, input_block: List[RotationGateMap], input_coeffs: List[float]
    ) -> None:

        """
        Splits the coefficients and gatemaps into qubit singles and qubit pairs.
        """

        single_qubit_coeffs = []
        pair_qubit_coeffs = []
        qubit_singles = []
        qubit_pairs = []

        if len(input_block) != len(input_coeffs):

            raise ValueError(
                "The number of terms/gatemaps must match the number of coefficients provided."
            )

        for each_gatemap, each_coeff in zip(input_block, input_coeffs):

            if each_gatemap.pauli_label[0] == "1q":
                single_qubit_coeffs.append(each_coeff)
                # Giving a string name to each gatemap (?)
                qubit_singles.append(type(each_gatemap).__name__)
            elif each_gatemap.pauli_label[0] == "2q":
                pair_qubit_coeffs.append(each_coeff)
                qubit_pairs.append(type(each_gatemap).__name__)

        return (single_qubit_coeffs, pair_qubit_coeffs, qubit_singles, qubit_pairs)

    @property
    def cost_block(self):

        return self._cost_block

    @cost_block.setter
    def cost_block(self, input_object: Hamiltonian) -> None:

        self._cost_block = HamiltonianMapper.repeat_gate_maps(
            input_object, "cost", self.p
        )

    @property
    def mixer_block(self):

        return self._mixer_block

    @mixer_block.setter
    def mixer_block(
        self, input_object: Union[List[RotationGateMap], Hamiltonian]
    ) -> None:

        if type(input_object) == Hamiltonian:

            self._mixer_block = HamiltonianMapper.repeat_gate_maps(
                input_object, "mixer", self.p
            )

        else:

            self._mixer_block = HamiltonianMapper.repeat_gate_maps_from_gate_map_list(
                input_object, "mixer", self.p
            )

    def __repr__(self):

        """Return an overview over the parameters and hyperparameters

        Todo
        ----
        Split this into ``__repr__`` and ``__str__`` with a more verbose
        output in ``__repr__``.
        """

        string = "Circuit Parameters:\n"
        string += "\tp: " + str(self.p) + "\n"
        string += "\tregister: " + str(self.qureg) + "\n" + "\n"

        string += "Cost Hamiltonian:\n"
        string += "\tcost_qubits_singles: " + str(self.cost_qubits_singles) + "\n"
        string += (
            "\tcost_single_qubit_coeffs: " + str(self.cost_single_qubit_coeffs) + "\n"
        )
        string += "\tcost_qubits_pairs: " + str(self.cost_qubits_pairs) + "\n"
        string += (
            "\tcost_pair_qubit_coeffs: "
            + str(self.cost_pair_qubit_coeffs)
            + "\n"
            + "\n"
        )

        string += "Mixer Hamiltonian:\n"
        string += "\tmixer_qubits_singles: " + str(self.mixer_qubits_singles) + "\n"
        string += (
            "\tmixer_single_qubit_coeffs: " + str(self.mixer_single_qubit_coeffs) + "\n"
        )
        string += "\tmixer_qubits_pairs: " + str(self.mixer_qubits_pairs) + "\n"
        string += (
            "\tmixer_pair_qubit_coeffs: " + str(self.mixer_pair_qubit_coeffs) + "\n"
        )

        return string

    @property
    def abstract_circuit(self):

        _abstract_circuit = []
        for each_p in range(self.p):
            _abstract_circuit.extend(self.cost_block[each_p])
            _abstract_circuit.extend(self.mixer_block[each_p])

        return _abstract_circuit



import unittest

from openqaoa.optimizers.logger_vqa import (
    Logger, LoggerVariable, LoggerVariableFactory,
    EmptyValue, AppendValue, ReplaceValue,
    IfLowerDo, IfHigherDo
)

class TestingLoggerClass(unittest.TestCase):
    
    def test_logger_var_setting_1(self):
        
        test_var = LoggerVariable('test_var', EmptyValue, EmptyValue)
        
        test_var.update(10)
        
        self.assertEqual(test_var.history, [])
        self.assertEqual(test_var.best, [])
        
    def test_logger_var_setting_2(self):
        
        test_var = LoggerVariable('test_var', AppendValue, AppendValue)
        
        test_var.update(5)
        test_var.update(10)
        
        self.assertEqual(test_var.history, [5, 10])
        self.assertEqual(test_var.best, [5, 10])
        
    def test_logger_var_setting_3(self):
        
        test_var = LoggerVariable('test_var', ReplaceValue, ReplaceValue)
        
        test_var.update(5)
        test_var.update(10)
        
        self.assertEqual(test_var.history, [10])
        self.assertEqual(test_var.best, [10])
        
    def test_logger_var_setting_4(self):
        
        test_var = LoggerVariable('test_var', IfLowerDo(EmptyValue), IfLowerDo(EmptyValue))
        
        test_var.update(5) # First value gets appended regardless
        test_var.update(10)
        
        self.assertEqual(test_var.history, [5])
        self.assertEqual(test_var.best, [5])
        
        test_var.update(1)
        
        self.assertEqual(test_var.history, [])
        self.assertEqual(test_var.best, [])
        
    def test_logger_var_setting_5(self):
        
        test_var = LoggerVariable('test_var', IfLowerDo(AppendValue), IfLowerDo(AppendValue))
        
        test_var.update(5) # First value gets appended regardless
        test_var.update(10)
        
        self.assertEqual(test_var.history, [5])
        self.assertEqual(test_var.best, [5])
        
        test_var.update(1)
        
        self.assertEqual(test_var.history, [5, 1])
        self.assertEqual(test_var.best, [5, 1])
        
    def test_logger_var_setting_6(self):
        
        test_var = LoggerVariable('test_var', IfLowerDo(ReplaceValue), IfLowerDo(ReplaceValue))
        
        test_var.update(5) # First value gets appended regardless
        test_var.update(10)
        
        self.assertEqual(test_var.history, [5])
        self.assertEqual(test_var.best, [5])
        
        test_var.update(1)
        
        self.assertEqual(test_var.history, [1])
        self.assertEqual(test_var.best, [1])
        
    def test_logger_var_setting_7(self):
        
        test_var = LoggerVariable('test_var', IfHigherDo(EmptyValue), IfHigherDo(EmptyValue))
        
        test_var.update(5) # First value gets appended regardless
        test_var.update(1)
        
        self.assertEqual(test_var.history, [5])
        self.assertEqual(test_var.best, [5])
        
        test_var.update(10)
        
        self.assertEqual(test_var.history, [])
        self.assertEqual(test_var.best, [])
        
    def test_logger_var_setting_8(self):
        
        test_var = LoggerVariable('test_var', IfHigherDo(AppendValue), IfHigherDo(AppendValue))
        
        test_var.update(5) # First value gets appended regardless
        test_var.update(1)
        
        self.assertEqual(test_var.history, [5])
        self.assertEqual(test_var.best, [5])
        
        test_var.update(10)
        
        self.assertEqual(test_var.history, [5, 10])
        self.assertEqual(test_var.best, [5, 10])
        
    def test_logger_var_setting_9(self):
        
        test_var = LoggerVariable('test_var', IfHigherDo(ReplaceValue), IfHigherDo(ReplaceValue))
        
        test_var.update(5) # First value gets appended regardless
        test_var.update(1)
        
        self.assertEqual(test_var.history, [5])
        self.assertEqual(test_var.best, [5])
        
        test_var.update(10)
        
        self.assertEqual(test_var.history, [10])
        self.assertEqual(test_var.best, [10])
        
    def test_logger_var_methods(self):
        
        test_var = LoggerVariable('test_var', ReplaceValue, ReplaceValue)
        test_var_2 = LoggerVariable('test_var', ReplaceValue, ReplaceValue)
        
        test_var.update(1)
        
        test_var_2.update_history(1)
        test_var_2.update_best(1)
        
        self.assertEqual(test_var.history, test_var_2.history)
        self.assertEqual(test_var.best, test_var_2.best)
        
    def test_logger_var_fact_hist_bool(self):
        
        var_1 = LoggerVariableFactory.create_logger_variable('new_attribute', True, 'Replace')
        var_2 = LoggerVariableFactory.create_logger_variable('new_attribute', False, 'Replace')
        
        var_1.update(1)
        var_2.update(1)
        
        var_1.update(2)
        var_2.update(2)
        
        self.assertEqual(var_1.name, 'new_attribute')
        self.assertEqual(var_2.name, 'new_attribute')
        self.assertEqual(var_1.history, [1, 2])
        self.assertEqual(var_2.history, [])
        
    def test_logger_var_fact_best_str(self):
        
        """
        Testing all possible settings of a variable created through the Factory
        Class
        """
        
        var_1 = LoggerVariableFactory.create_logger_variable('new_attribute', True, 'Append')
        var_2 = LoggerVariableFactory.create_logger_variable('new_attribute', True, 'Replace')
        var_3 = LoggerVariableFactory.create_logger_variable('new_attribute', True, 'LowestSoFar')
        var_4 = LoggerVariableFactory.create_logger_variable('new_attribute', True, 'LowestOnly')
        var_5 = LoggerVariableFactory.create_logger_variable('new_attribute', True, 'HighestSoFar')
        var_6 = LoggerVariableFactory.create_logger_variable('new_attribute', True, 'HighestOnly')
        
        values_list = [5, 7, 10, 1]
        
        for each_value in values_list:
            var_1.update(each_value)
            var_2.update(each_value)
            var_3.update(each_value)
            var_4.update(each_value)
            var_5.update(each_value)
            var_6.update(each_value)
            
        self.assertEqual(var_1.best, [5, 7, 10, 1])
        self.assertEqual(var_2.best, [1])
        self.assertEqual(var_3.best, [5, 1])
        self.assertEqual(var_4.best, [1])
        self.assertEqual(var_5.best, [5, 7, 10])
        self.assertEqual(var_6.best, [10])
        
    def test_logger_obj_update_struct_1(self):
        
        """
        The best update structure ensures that the best value for that particular
        attribute respects a relation with another attribute.
        
        In the code below, the following can be observed:
        Even though the lowest in attribute 2 is 10. The best of attribute 2 is
        15 instead. This is because the best value of attribute 2 depends on whether
        attribute 1 for that set of update values is the highest. If it is not the
        highest value, attribute 2's best value is not updated. This way updating
        the best value of attribute 2 respects the update on the best value of 
        attribute 1.
        
        Histories are always updated Indepedent of the best update structure provided.
        """
        
        logger_obj = Logger({'attribute_1': 
                            {
                                'history_update_bool': True, 
                                'best_update_string': 'HighestOnly'
                            }, 
                            'attribute_2': 
                            {
                                'history_update_bool': True, 
                                'best_update_string': 'LowestSoFar'
                            },
                            'attribute_3': 
                            {
                                'history_update_bool': False, 
                                'best_update_string': 'Replace'
                            }
                            }, 
                            {
                                'root_nodes': ['attribute_1'],
                                'best_update_structure': 
                                (['attribute_1', 'attribute_2'], 
                                 ['attribute_1', 'attribute_3'])
                            })
        
        logger_obj.log_variables({'attribute_1': 10, 
                                  'attribute_2': 3, 
                                  'attribute_3': 'string 1'})
        
        self.assertEqual(logger_obj.attribute_1.history, [10])
        self.assertEqual(logger_obj.attribute_2.history, [3])
        self.assertEqual(logger_obj.attribute_3.history, [])
        
        self.assertEqual(logger_obj.attribute_1.best, [10])
        self.assertEqual(logger_obj.attribute_2.best, [3])
        self.assertEqual(logger_obj.attribute_3.best, ['string 1'])
        
        logger_obj.log_variables({'attribute_1': 5, 
                                  'attribute_2': 1, 
                                  'attribute_3': 'string 2'})
        
        self.assertEqual(logger_obj.attribute_1.history, [10, 5])
        self.assertEqual(logger_obj.attribute_2.history, [3, 1])
        self.assertEqual(logger_obj.attribute_3.history, [])
        
        self.assertEqual(logger_obj.attribute_1.best, [10])
        self.assertEqual(logger_obj.attribute_2.best, [3])
        self.assertEqual(logger_obj.attribute_3.best, ['string 1'])
        
        logger_obj.log_variables({'attribute_1': 15, 
                                  'attribute_2': 4, 
                                  'attribute_3': 'string 3'})
        
        self.assertEqual(logger_obj.attribute_1.history, [10, 5, 15])
        self.assertEqual(logger_obj.attribute_2.history, [3, 1, 4])
        self.assertEqual(logger_obj.attribute_3.history, [])
        
        self.assertEqual(logger_obj.attribute_1.best, [15])
        self.assertEqual(logger_obj.attribute_2.best, [3])
        self.assertEqual(logger_obj.attribute_3.best, ['string 3'])
        
        logger_obj.log_variables({'attribute_1': 20, 
                                  'attribute_2': 2, 
                                  'attribute_3': 'string 4'})
        
        self.assertEqual(logger_obj.attribute_1.history, [10, 5, 15, 20])
        self.assertEqual(logger_obj.attribute_2.history, [3, 1, 4, 2])
        self.assertEqual(logger_obj.attribute_3.history, [])
        
        self.assertEqual(logger_obj.attribute_1.best, [20])
        self.assertEqual(logger_obj.attribute_2.best, [3, 2])
        self.assertEqual(logger_obj.attribute_3.best, ['string 4'])
        
    def test_logger_obj_update_struct_2(self):
        
        """
        Single layer test
        """
        
        logger_obj = Logger({'attribute_1': 
                            {
                                'history_update_bool': True, 
                                'best_update_string': 'HighestOnly'
                            }, 
                            'attribute_2': 
                            {
                                'history_update_bool': True, 
                                'best_update_string': 'LowestSoFar'
                            },
                            'attribute_3': 
                            {
                                'history_update_bool': False, 
                                'best_update_string': 'Replace'
                            }
                            }, 
                            {
                                'root_nodes': ['attribute_1', 'attribute_2', 'attribute_3'],
                                'best_update_structure': ([])
                            })
        
        logger_obj.log_variables({'attribute_1': 10, 
                                  'attribute_2': 3, 
                                  'attribute_3': 'string 1'})
        
        self.assertEqual(logger_obj.attribute_1.history, [10])
        self.assertEqual(logger_obj.attribute_2.history, [3])
        self.assertEqual(logger_obj.attribute_3.history, [])
        
        self.assertEqual(logger_obj.attribute_1.best, [10])
        self.assertEqual(logger_obj.attribute_2.best, [3])
        self.assertEqual(logger_obj.attribute_3.best, ['string 1'])
        
        logger_obj.log_variables({'attribute_1': 5, 
                                  'attribute_2': 1, 
                                  'attribute_3': 'string 2'})
        
        self.assertEqual(logger_obj.attribute_1.history, [10, 5])
        self.assertEqual(logger_obj.attribute_2.history, [3, 1])
        self.assertEqual(logger_obj.attribute_3.history, [])
        
        self.assertEqual(logger_obj.attribute_1.best, [10])
        self.assertEqual(logger_obj.attribute_2.best, [3, 1])
        self.assertEqual(logger_obj.attribute_3.best, ['string 2'])
        
        logger_obj.log_variables({'attribute_1': 15, 
                                  'attribute_2': 4, 
                                  'attribute_3': 'string 3'})
        
        self.assertEqual(logger_obj.attribute_1.history, [10, 5, 15])
        self.assertEqual(logger_obj.attribute_2.history, [3, 1, 4])
        self.assertEqual(logger_obj.attribute_3.history, [])
        
        self.assertEqual(logger_obj.attribute_1.best, [15])
        self.assertEqual(logger_obj.attribute_2.best, [3, 1])
        self.assertEqual(logger_obj.attribute_3.best, ['string 3'])
        
        logger_obj.log_variables({'attribute_1': 20, 
                                  'attribute_2': 2, 
                                  'attribute_3': 'string 4'})
        
        self.assertEqual(logger_obj.attribute_1.history, [10, 5, 15, 20])
        self.assertEqual(logger_obj.attribute_2.history, [3, 1, 4, 2])
        self.assertEqual(logger_obj.attribute_3.history, [])
        
        self.assertEqual(logger_obj.attribute_1.best, [20])
        self.assertEqual(logger_obj.attribute_2.best, [3, 1])
        self.assertEqual(logger_obj.attribute_3.best, ['string 4'])
        
    def test_logger_obj_update_struct_3(self):
        
        """
        Multi layer test
        """
        
        logger_obj = Logger({'attribute_1': 
                            {
                                'history_update_bool': True, 
                                'best_update_string': 'HighestOnly'
                            }, 
                            'attribute_2': 
                            {
                                'history_update_bool': True, 
                                'best_update_string': 'LowestSoFar'
                            },
                            'attribute_3': 
                            {
                                'history_update_bool': False, 
                                'best_update_string': 'Replace'
                            }
                            }, 
                            {
                                'root_nodes': ['attribute_1'],
                                'best_update_structure': 
                                (['attribute_1', 'attribute_2'], 
                                 ['attribute_2', 'attribute_3'])
                            })
        
        logger_obj.log_variables({'attribute_1': 10, 
                                  'attribute_2': 3, 
                                  'attribute_3': 'string 1'})
        
        self.assertEqual(logger_obj.attribute_1.history, [10])
        self.assertEqual(logger_obj.attribute_2.history, [3])
        self.assertEqual(logger_obj.attribute_3.history, [])
        
        self.assertEqual(logger_obj.attribute_1.best, [10])
        self.assertEqual(logger_obj.attribute_2.best, [3])
        self.assertEqual(logger_obj.attribute_3.best, ['string 1'])
        
        logger_obj.log_variables({'attribute_1': 5, 
                                  'attribute_2': 1, 
                                  'attribute_3': 'string 2'})
        
        self.assertEqual(logger_obj.attribute_1.history, [10, 5])
        self.assertEqual(logger_obj.attribute_2.history, [3, 1])
        self.assertEqual(logger_obj.attribute_3.history, [])
        
        self.assertEqual(logger_obj.attribute_1.best, [10])
        self.assertEqual(logger_obj.attribute_2.best, [3])
        self.assertEqual(logger_obj.attribute_3.best, ['string 1'])
        
        logger_obj.log_variables({'attribute_1': 15, 
                                  'attribute_2': 4, 
                                  'attribute_3': 'string 3'})
        
        self.assertEqual(logger_obj.attribute_1.history, [10, 5, 15])
        self.assertEqual(logger_obj.attribute_2.history, [3, 1, 4])
        self.assertEqual(logger_obj.attribute_3.history, [])
        
        self.assertEqual(logger_obj.attribute_1.best, [15])
        self.assertEqual(logger_obj.attribute_2.best, [3])
        self.assertEqual(logger_obj.attribute_3.best, ['string 1'])
        
        logger_obj.log_variables({'attribute_1': 20, 
                                  'attribute_2': 2, 
                                  'attribute_3': 'string 4'})
        
        self.assertEqual(logger_obj.attribute_1.history, [10, 5, 15, 20])
        self.assertEqual(logger_obj.attribute_2.history, [3, 1, 4, 2])
        self.assertEqual(logger_obj.attribute_3.history, [])
        
        self.assertEqual(logger_obj.attribute_1.best, [20])
        self.assertEqual(logger_obj.attribute_2.best, [3, 2])
        self.assertEqual(logger_obj.attribute_3.best, ['string 4'])

if __name__ == '__main__':
    unittest.main()


"""
Test the plot functionalities in qaoa_parameters 
"""

# import the standard modules from python
import unittest

# import the OpenQAOA Parameterisation classes manually: Manual Mode
from openqaoa.qaoa_components import (
    create_qaoa_variational_params, PauliOp,
    Hamiltonian, QAOADescriptor
)

# import the other OpenQAOA modules required for this example
from openqaoa.utilities import X_mixer_hamiltonian
from openqaoa.problems import MaximumCut


# Create a hamiltonian problem on 15 qubits and create Circuit Params Class
hamiltonian1 = MaximumCut.random_instance(n_nodes=15, edge_probability = 0.9).qubo.hamiltonian 
mixer_hamiltonian1 = X_mixer_hamiltonian(n_qubits = 15)
qaoa_descriptor1 = QAOADescriptor(cost_hamiltonian = hamiltonian1, mixer_block = mixer_hamiltonian1, p=4)

# Create a hamiltonian on 3 qubits with 2 coupling terms and 1 bias term and create Circuit Params Class
Term1 = PauliOp('ZZ', (0, 1)) 
Term2 = PauliOp('ZZ', (0, 2))
Term3 = PauliOp('Z', (0, ))

hamiltonian2 = Hamiltonian([Term1, Term2, Term3], [0.7, 1.2, -0.5], 0.0)
mixer_hamiltonian2 = X_mixer_hamiltonian(n_qubits = 3)
qaoa_descriptor2 = QAOADescriptor(cost_hamiltonian = hamiltonian2, mixer_block = mixer_hamiltonian2, p=4)


class TestingQAOAParametersPlots(unittest.TestCase):

    def _plot(self, qaoa_descriptor, params_type, init_type = 'rand'):
        #generic function to create the plots
        print(f'Example of {params_type} params')
        if params_type in ['fourier', 'fourier_extended', 'fourier_w_bias']:
            params = create_qaoa_variational_params(qaoa_descriptor = qaoa_descriptor, params_type = params_type, init_type = init_type, q=2)
        else:
            params = create_qaoa_variational_params(qaoa_descriptor = qaoa_descriptor, params_type = params_type, init_type = init_type)
        params.plot()

    def test_QAOAVariationalStandardParamsPlots(self):        # standard params
        self._plot(qaoa_descriptor1, 'standard')

    def test_QAOAVariationalStandardBiasParamsPlots(self):    # standard_w_bias params
        self._plot(qaoa_descriptor2, 'standard_w_bias')

    def test_QAOAVariationalExtendedParamsPlots(self):        # extended params
        self._plot(qaoa_descriptor1, 'extended')
        self._plot(qaoa_descriptor2, 'extended')    
        
    def test_QAOAVariationalFourierParamsPlots(self):        # fourier params
        self._plot(qaoa_descriptor2, 'fourier')    
        
    def test_QAOAVariationalFourierBiasParamsPlots(self):        # fourier_w_bias params
        self._plot(qaoa_descriptor2, 'fourier_w_bias')  
        
    def test_QAOAVariationalFourierExtendedParamsPlots(self):        # fourier_extended params
        self._plot(qaoa_descriptor1, 'fourier_extended')
        self._plot(qaoa_descriptor2, 'fourier_extended')   
        
    def test_QAOAVariationalAnnealingParamsPlots(self):        # annealing params
        self._plot(qaoa_descriptor2, 'annealing', init_type = 'ramp')  



if __name__ == "__main__":
    unittest.main()




import unittest
import importlib

from setuptools import find_namespace_packages

class TestImports(unittest.TestCase):

    """
    Test module imports within OpenQAOA
    """

    def test_all_module_import(self):
        """
        Test all the main module imports for OQ
        """
        
        package_names = ['openqaoa', 'openqaoa_braket', 'openqaoa_qiskit', 'openqaoa_pyquil', 'openqaoa_azure']
        folder_names = ['openqaoa-core', 'openqaoa-braket', 'openqaoa-qiskit', 'openqaoa-pyquil', 'openqaoa-azure']
        packages_import = find_namespace_packages(where="./src")
        updated_packages = []
        for each_package_name in packages_import:
            for _index, each_folder_name in enumerate(folder_names):
                if each_folder_name in each_package_name:
                    updated_packages.append(each_package_name.replace(each_folder_name, package_names[_index]))
                    continue

        for each_package in updated_packages:
            try:
                importlib.import_module(each_package)
            except Exception as e:
                raise Exception(e)


if __name__ == "__main__":
    unittest.main()
